"""
Module for managing keyboards on supported POSIX-like systems using
systemd, or such as Redhat, Debian and Gentoo.
"""

import logging

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only works with systemd or on supported POSIX-like systems
    """
    if salt.utils.path.which("localectl") or __grains__["os_family"] in (
        "RedHat",
        "Debian",
        "Gentoo",
    ):
        return True
    return (
        False,
        "The keyboard exeuction module cannot be loaded: "
        "only works on Redhat, Debian or Gentoo systems or if localectl binary in"
        " path.",
    )


def get_sys():
    """
    Get current system keyboard setting

    CLI Example:

    .. code-block:: bash

        salt '*' keyboard.get_sys
    """
    cmd = ""
    if salt.utils.path.which("localectl"):
        cmd = 'localectl | grep Keymap | sed -e"s/: /=/" -e"s/^[ \t]*//"'
    elif "RedHat" in __grains__["os_family"]:
        cmd = 'grep LAYOUT /etc/sysconfig/keyboard | grep -vE "^#"'
    elif "Debian" in __grains__["os_family"]:
        cmd = 'grep XKBLAYOUT /etc/default/keyboard | grep -vE "^#"'
    elif "Gentoo" in __grains__["os_family"]:
        cmd = 'grep "^keymap" /etc/conf.d/keymaps | grep -vE "^#"'
    out = __salt__["cmd.run"](cmd, python_shell=True).split("=")
    ret = out[1].replace('"', "")
    return ret


def set_sys(layout):
    """
    Set current system keyboard setting

    CLI Example:

    .. code-block:: bash

        salt '*' keyboard.set_sys dvorak
    """
    if salt.utils.path.which("localectl"):
        __salt__["cmd.run"](f"localectl set-keymap {layout}")
    elif "RedHat" in __grains__["os_family"]:
        __salt__["file.sed"](
            "/etc/sysconfig/keyboard", "^LAYOUT=.*", f"LAYOUT={layout}"
        )
    elif "Debian" in __grains__["os_family"]:
        __salt__["file.sed"](
            "/etc/default/keyboard", "^XKBLAYOUT=.*", f"XKBLAYOUT={layout}"
        )
    elif "Gentoo" in __grains__["os_family"]:
        __salt__["file.sed"]("/etc/conf.d/keymaps", "^keymap=.*", f"keymap={layout}")
    return layout


def get_x():
    """
    Get current X keyboard setting

    CLI Example:

    .. code-block:: bash

        salt '*' keyboard.get_x
    """
    cmd = "setxkbmap -query | grep layout"
    out = __salt__["cmd.run"](cmd, python_shell=True).split(":")
    return out[1].strip()


def set_x(layout):
    """
    Set current X keyboard setting

    CLI Example:

    .. code-block:: bash

        salt '*' keyboard.set_x dvorak
    """
    cmd = f"setxkbmap {layout}"
    __salt__["cmd.run"](cmd)
    return layout



import pathlib
import shutil
import threading
import time

import pytest

from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]

SSH_SLS = "ssh_state_tests"
SSH_SLS_ID = "ssh-file-test"
SSH_SLS_FILE = pathlib.Path("/tmp/salt_test_file")


@pytest.fixture(autouse=True)
def cleanup_thin_dir(salt_ssh_cli):
    """
    Ensure the thin_dir and any files created by these tests are removed.
    """
    try:
        yield
    finally:
        ret = salt_ssh_cli.run("config.get", "thin_dir")
        if ret.returncode == 0 and ret.data:
            shutil.rmtree(ret.data, ignore_errors=True)
        for path in SSH_SLS_FILE.parent.glob(f"{SSH_SLS_FILE.name}*"):
            path.unlink(missing_ok=True)


def _assert_state_dict(ret):
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    return ret.data


def test_state_apply(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.apply", SSH_SLS)
    data = _assert_state_dict(ret)
    assert all(item.get("__sls__") == SSH_SLS for item in data.values())

    exists = salt_ssh_cli.run("file.file_exists", str(SSH_SLS_FILE))
    assert exists.returncode == 0
    assert exists.data is True


def test_state_sls_id_test_mode(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.sls_id", SSH_SLS_ID, SSH_SLS, "test=True")
    data = _assert_state_dict(ret)
    comment = next(iter(data.values())).get("comment", "")
    assert "No changes made" in comment


def test_state_sls_id(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.sls_id", SSH_SLS_ID, SSH_SLS)
    data = _assert_state_dict(ret)
    ids = {item.get("__id__") for item in data.values()}
    assert ids == {SSH_SLS_ID}

    exists = salt_ssh_cli.run("file.file_exists", str(SSH_SLS_FILE))
    assert exists.returncode == 0
    assert exists.data is True


def test_state_sls_wrong_id(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.sls_id", "doesnotexist", SSH_SLS)
    assert "No matches for ID" in ret.stdout


def test_state_sls_id_with_pillar(salt_ssh_cli):
    ret = salt_ssh_cli.run(
        "state.sls_id",
        SSH_SLS_ID,
        SSH_SLS,
        pillar='{"test_file_suffix": "_pillar"}',
    )
    _assert_state_dict(ret)

    pillar_file = SSH_SLS_FILE.with_name(SSH_SLS_FILE.name + "_pillar")
    exists = salt_ssh_cli.run("file.file_exists", str(pillar_file))
    assert exists.returncode == 0
    assert exists.data is True


def test_state_show_sls(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.show_sls", SSH_SLS)
    data = _assert_state_dict(ret)
    assert all(item.get("__sls__") == SSH_SLS for item in data.values())


def test_state_sls_exists(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.sls_exists", SSH_SLS)
    assert ret.returncode == 0
    assert ret.data is True


def test_state_show_top(salt_ssh_cli, base_env_state_tree_root_dir):
    top_sls = """
    base:
      '*':
        - core
    """
    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
    """.format(
        RUNTIME_VARS.TMP
    )
    with pytest.helpers.temp_file(
        "top.sls", top_sls, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file("core.sls", core_state, base_env_state_tree_root_dir):
        ret = salt_ssh_cli.run("state.show_top")
        assert ret.returncode == 0
        assert ret.data == {"base": ["core", "master_tops_test"]}


def test_state_single(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.single", "test.succeed_with_changes", "name=itworked")
    data = _assert_state_dict(ret)
    state_res = next(iter(data.values()))
    assert state_res["name"] == "itworked"
    assert state_res["result"] is True
    assert state_res["comment"] == "Success!"


def test_state_show_highstate(salt_ssh_cli, base_env_state_tree_root_dir):
    top_sls = """
    base:
      '*':
        - core
    """
    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
    """.format(
        RUNTIME_VARS.TMP
    )
    with pytest.helpers.temp_file(
        "top.sls", top_sls, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file("core.sls", core_state, base_env_state_tree_root_dir):
        ret = salt_ssh_cli.run("state.show_highstate")
        data = _assert_state_dict(ret)
        dest = f"{RUNTIME_VARS.TMP}/testfile"
        assert dest in data
        assert data[dest]["__env__"] == "base"


def test_state_high(salt_ssh_cli):
    ret = salt_ssh_cli.run(
        "state.high", '{"itworked": {"test": ["succeed_with_changes"]}}'
    )
    data = _assert_state_dict(ret)
    state_res = next(iter(data.values()))
    assert state_res["name"] == "itworked"
    assert state_res["result"] is True
    assert state_res["comment"] == "Success!"


def test_state_show_lowstate(salt_ssh_cli, base_env_state_tree_root_dir):
    top_sls = """
    base:
      '*':
        - core
    """
    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
    """.format(
        RUNTIME_VARS.TMP
    )
    with pytest.helpers.temp_file(
        "top.sls", top_sls, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file("core.sls", core_state, base_env_state_tree_root_dir):
        ret = salt_ssh_cli.run("state.show_lowstate")
        assert ret.returncode == 0
        assert isinstance(ret.data, list)
        assert ret.data
        assert isinstance(ret.data[0], dict)


def test_state_low(salt_ssh_cli):
    ret = salt_ssh_cli.run(
        "state.low",
        '{"state": "test", "fun": "succeed_with_changes", "name": "itworked"}',
    )
    data = _assert_state_dict(ret)
    state_res = next(iter(data.values()))
    assert state_res["name"] == "itworked"
    assert state_res["result"] is True
    assert state_res["comment"] == "Success!"


def test_state_request_check_clear(salt_ssh_cli):
    request = salt_ssh_cli.run("state.request", SSH_SLS)
    _assert_state_dict(request)

    check = salt_ssh_cli.run("state.check_request")
    assert check.returncode == 0
    assert check.data

    clear = salt_ssh_cli.run("state.clear_request")
    assert clear.returncode == 0

    check_empty = salt_ssh_cli.run("state.check_request")
    assert check_empty.returncode == 0
    assert not check_empty.data


def test_state_run_request(salt_ssh_cli):
    request = salt_ssh_cli.run("state.request", SSH_SLS)
    _assert_state_dict(request)

    run = salt_ssh_cli.run("state.run_request")
    _assert_state_dict(run)

    exists = salt_ssh_cli.run("file.file_exists", str(SSH_SLS_FILE))
    assert exists.returncode == 0
    assert exists.data is True


def test_state_running(
    salt_master, salt_ssh_cli, salt_ssh_roster_file, sshd_config_dir
):
    results = []
    background_cli = salt_master.salt_ssh_cli(
        timeout=180,
        roster_file=salt_ssh_roster_file,
        target_host="localhost",
        client_key=str(sshd_config_dir / "client_key"),
    )

    def _run_state():
        results.append(background_cli.run("state.sls", "running"))

    thread = threading.Thread(target=_run_state)
    thread.start()

    expected = 'The function "state.pkg" is running as'
    try:
        end_time = time.time() + 30
        while time.time() < end_time:
            ret = salt_ssh_cli.run("state.running")
            if isinstance(ret.data, list):
                output = " ".join(ret.data)
            else:
                output = " ".join(str(ret.data).splitlines())
            if expected in output:
                break
            time.sleep(1)
        else:
            if results and "Failed to return clean data" in str(results[0].data):
                pytest.skip("Background state run failed, skipping")
            pytest.fail(f"Did not find '{expected}' in state.running output")
    finally:
        thread.join(timeout=120)

    end_time = time.time() + 120
    while time.time() < end_time:
        ret = salt_ssh_cli.run("state.running")
        if isinstance(ret.data, list):
            output = " ".join(ret.data)
        else:
            output = " ".join(str(ret.data).splitlines())
        if expected not in output:
            break
        time.sleep(1)
    else:
        pytest.fail("state.pkg is still reported as running")



"""
Functional tests for chocolatey state
"""

import os
import pathlib

import pytest

import salt.utils.path
import salt.utils.win_reg

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
]


@pytest.fixture(scope="module")
def chocolatey(states):
    yield states.chocolatey


@pytest.fixture(scope="module")
def chocolatey_mod(modules):

    current_path = salt.utils.win_reg.read_value(
        hive="HKLM",
        key=r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment",
        vname="PATH",
    )["vdata"]
    url = "https://community.chocolatey.org/api/v2/package/chocolatey/"
    with pytest.helpers.temp_file(name="choco.nupkg") as nupkg:
        choco_pkg = pathlib.Path(str(nupkg))
    choco_dir = choco_pkg.parent / "choco_dir"
    choco_script = choco_dir / "tools" / "chocolateyInstall.ps1"

    def install():
        # Install Chocolatey 1.2.1

        # Download Package
        modules.cp.get_url(path=url, dest=str(choco_pkg))

        # Unzip Package
        modules.archive.unzip(
            zip_file=str(choco_pkg),
            dest=str(choco_dir),
            extract_perms=False,
        )

        # Run installer script
        assert choco_script.exists()
        result = modules.cmd.script(
            source=str(choco_script),
            cwd=str(choco_script.parent),
            shell="powershell",
            python_shell=True,
        )
        assert result["retcode"] == 0

    def uninstall():
        choco_dir = os.environ.get("ChocolateyInstall", False)
        if choco_dir:
            # Remove Chocolatey Directory
            modules.file.remove(path=choco_dir, force=True)
            # Remove Chocolatey Environment Variables
            for env_var in modules.environ.items():
                if env_var.lower().startswith("chocolatey"):
                    modules.environ.setval(
                        key=env_var, val=False, false_unsets=True, permanent="HKLM"
                    )
                    modules.environ.setval(
                        key=env_var, val=False, false_unsets=True, permanent="HKCU"
                    )
        salt.utils.win_reg.set_value(
            hive="HKLM",
            key=r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment",
            vname="PATH",
            vdata=current_path,
        )
        modules.win_path.rehash()

    # Remove unknown version
    if salt.utils.path.which("choco.exe"):
        uninstall()

    # Install known version
    install()

    yield modules.chocolatey

    # Remove
    uninstall()


@pytest.fixture(scope="function")
def clean(chocolatey_mod):
    chocolatey_mod.uninstall(name="vim", force=True)
    yield
    chocolatey_mod.uninstall(name="vim", force=True)


@pytest.fixture(scope="function")
def vim(chocolatey_mod):
    chocolatey_mod.install(name="vim", version="9.0.1672")
    yield
    chocolatey_mod.uninstall(name="vim", force=True)


@pytest.fixture(scope="function")
def sudo(chocolatey_mod):
    chocolatey_mod.install(name="sudo", version="1.1.2")
    yield
    chocolatey_mod.uninstall(name="sudo", force=True)


def test_installed_latest(clean, chocolatey, chocolatey_mod):
    chocolatey.installed(name="vim")
    result = chocolatey_mod.version(name="vim")
    assert "vim" in result


def test_installed_version(clean, chocolatey, chocolatey_mod):
    chocolatey.installed(name="vim", version="9.0.1672")
    result = chocolatey_mod.version(name="vim")
    assert "vim" in result
    assert result["vim"]["installed"][0] == "9.0.1672"


# @pytest.mark.skipif(True, reason="Timing out, skipping for now")
def test_installed_version_existing_capitalization(sudo, chocolatey, chocolatey_mod):
    result = chocolatey.installed(name="sudo", version="1.1.3")
    expected_changes = {"Sudo": {"new": ["1.1.3"], "old": ["1.1.2"]}}
    assert result["changes"] == expected_changes


def test_uninstalled(vim, chocolatey, chocolatey_mod):
    chocolatey.uninstalled(name="vim")
    result = chocolatey_mod.version(name="vim")
    assert "vim" not in result


def test_upgraded(vim, chocolatey, chocolatey_mod):
    result = chocolatey_mod.version(name="vim")
    assert "vim" in result
    assert result["vim"]["installed"][0] == "9.0.1672"
    chocolatey.upgraded(name="vim", version="9.0.1677")
    result = chocolatey_mod.version(name="vim")
    assert "vim" in result
    assert result["vim"]["installed"][0] == "9.0.1677"



import secrets

import pytest

import salt.states.file as file
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def configure_loader_modules():
    return {
        file: {"__opts__": {"test": False}},
    }


def test_cached_test_true():
    name = "salt://test/file.exe"
    source_hash = secrets.token_hex(nbytes=32)
    expected = {
        "changes": {},
        "comment": f"File will be cached: {name}",
        "name": name,
        "result": None,
    }
    salt = {
        "cp.is_cached": MagicMock(return_value=""),
        "file.get_source_sum": MagicMock(return_value={"hsum": source_hash}),
    }
    opts = {"test": True}
    with patch.dict(file.__salt__, salt), patch.dict(file.__opts__, opts):
        result = file.cached(name=name, source_hash=source_hash)
    assert result == expected


def test_cached_present_test_true():
    name = "salt://test/file.exe"
    source_hash = secrets.token_hex(nbytes=32)
    expected = {
        "changes": {},
        "comment": f"File already cached: {name}",
        "name": name,
        "result": None,
    }
    salt = {
        "cp.is_cached": MagicMock(return_value="path/to/file"),
        "file.get_hash": MagicMock(return_value=source_hash),
        "file.get_source_sum": MagicMock(return_value={"hsum": source_hash}),
    }
    opts = {"test": True, "hash_type": "sha256"}
    with patch.dict(file.__salt__, salt), patch.dict(file.__opts__, opts):
        result = file.cached(name=name, source_hash=source_hash)
    assert result == expected


def test_cached_present_different_hash_test_true():
    name = "salt://test/file.exe"
    source_hash = secrets.token_hex(nbytes=32)
    existing_hash = secrets.token_hex(nbytes=32)
    expected = {
        "changes": {},
        "comment": f"Hashes don't match.\nFile will be cached: {name}",
        "name": name,
        "result": None,
    }
    salt = {
        "cp.is_cached": MagicMock(return_value="path/to/file"),
        "file.get_hash": MagicMock(return_value=existing_hash),
        "file.get_source_sum": MagicMock(return_value={"hsum": source_hash}),
    }
    opts = {"test": True, "hash_type": "sha256"}
    with patch.dict(file.__salt__, salt), patch.dict(file.__opts__, opts):
        result = file.cached(name=name, source_hash=source_hash)
    assert result == expected


def test_cached_present_no_source_hash_test_true():
    name = "salt://test/file.exe"
    existing_hash = secrets.token_hex(nbytes=32)
    expected = {
        "changes": {},
        "comment": f"No hash found. File will be cached: {name}",
        "name": name,
        "result": None,
    }
    salt = {
        "cp.is_cached": MagicMock(return_value="path/to/file"),
        "file.get_hash": MagicMock(return_value=existing_hash),
    }
    opts = {"test": True, "hash_type": "sha256"}
    with patch.dict(file.__salt__, salt), patch.dict(file.__opts__, opts):
        result = file.cached(name=name)
    assert result == expected



"""
    Test cases for salt.states.etcd_mod

    Note: No functional tests are required as of now, as all of the
    functional pieces are already tested in utils.test_etcd_utils
    If the contents of this state were to add more logic besides
    essentially acting as a wrapper, then functional tests would be required.

    :codeauthor: Caleb Beard <calebb@vmware.com>
"""

import pytest

import salt.states.etcd_mod as etcd_state
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {etcd_state: {}}


def test_set():
    """
    Test the etcd_mod.set state function
    """
    get_mock = MagicMock()
    set_mock = MagicMock()
    dunder_salt = {
        "etcd.get": get_mock,
        "etcd.set": set_mock,
    }

    with patch.dict(etcd_state.__salt__, dunder_salt):
        # Test new key creation
        get_mock.return_value = None
        set_mock.return_value = "new value"
        expected = {
            "name": "new_key",
            "comment": "New key created",
            "result": True,
            "changes": {"new_key": "new value"},
        }
        assert etcd_state.set_("new_key", "new value", profile="test") == expected

        # Test key updating
        get_mock.return_value = "old value"
        set_mock.return_value = "new value"
        expected = {
            "name": "new_key",
            "comment": "Key value updated",
            "result": True,
            "changes": {"new_key": "new value"},
        }
        assert etcd_state.set_("new_key", "new value", profile="test") == expected

        # Test setting the same value to a key
        get_mock.return_value = "value"
        set_mock.return_value = "value"
        expected = {
            "name": "key",
            "comment": "Key contains correct value",
            "result": True,
            "changes": {},
        }
        assert etcd_state.set_("key", "value", profile="test") == expected


def test_wait_set():
    """
    Test the etcd_mod.wait_set state function
    """
    expected = {
        "name": "key",
        "changes": {},
        "result": True,
        "comment": "",
    }
    assert etcd_state.wait_set("key", "any value", profile="test") == expected


def test_directory():
    """
    Test the etcd_mod.directory state function
    """
    get_mock = MagicMock()
    set_mock = MagicMock()
    dunder_salt = {
        "etcd.get": get_mock,
        "etcd.set": set_mock,
    }

    with patch.dict(etcd_state.__salt__, dunder_salt):
        # Test new directory creation
        get_mock.return_value = None
        set_mock.return_value = "new_dir"
        expected = {
            "name": "new_dir",
            "comment": "New directory created",
            "result": True,
            "changes": {"new_dir": "Created"},
        }
        assert etcd_state.directory("new_dir", profile="test") == expected

        # Test creating an existing directory
        get_mock.return_value = "new_dir"
        set_mock.return_value = "new_dir"
        expected = {
            "name": "new_dir",
            "comment": "Directory exists",
            "result": True,
            "changes": {},
        }
        assert etcd_state.directory("new_dir", profile="test") == expected


def test_rm():
    """
    Test the etcd_mod.set state function
    """
    get_mock = MagicMock()
    rm_mock = MagicMock()
    dunder_salt = {
        "etcd.get": get_mock,
        "etcd.rm": rm_mock,
    }

    with patch.dict(etcd_state.__salt__, dunder_salt):
        # Test removing a key
        get_mock.return_value = "value"
        rm_mock.return_value = True
        expected = {
            "name": "key",
            "comment": "Key removed",
            "result": True,
            "changes": {"key": "Deleted"},
        }
        assert etcd_state.rm("key", profile="test") == expected

        # Test failing to remove an existing key
        get_mock.return_value = "value"
        rm_mock.return_value = False
        expected = {
            "name": "key",
            "comment": "Unable to remove key",
            "result": True,
            "changes": {},
        }
        assert etcd_state.rm("key", profile="test") == expected

        # Test removing a nonexistent key
        get_mock.return_value = False
        expected = {
            "name": "key",
            "comment": "Key does not exist",
            "result": True,
            "changes": {},
        }
        assert etcd_state.rm("key", profile="test") == expected


def test_wait_rm():
    """
    Test the etcd_mod.wait_rm state function
    """
    expected = {
        "name": "key",
        "changes": {},
        "result": True,
        "comment": "",
    }
    assert etcd_state.wait_rm("key", profile="test") == expected


def test_mod_watch():
    """
    Test the watch requisite function etcd_mod.mod_watch
    """
    get_mock = MagicMock()
    set_mock = MagicMock()
    rm_mock = MagicMock()
    dunder_salt = {
        "etcd.get": get_mock,
        "etcd.set": set_mock,
        "etcd.rm": rm_mock,
    }

    with patch.dict(etcd_state.__salt__, dunder_salt):
        # Test watch with wait_set
        get_mock.return_value = None
        set_mock.return_value = "value"
        expected = {
            "name": "key",
            "comment": "New key created",
            "result": True,
            "changes": {"key": "value"},
        }
        assert (
            etcd_state.mod_watch("key", value="value", sfun="wait_set", profile={})
            == expected
        )
        assert (
            etcd_state.mod_watch("key", value="value", sfun="wait_set_key", profile={})
            == expected
        )

        # Test watch with wait_rm
        get_mock.return_value = "value"
        rm_mock.return_value = True
        expected = {
            "name": "key",
            "comment": "Key removed",
            "result": True,
            "changes": {"key": "Deleted"},
        }
        assert etcd_state.mod_watch("key", sfun="wait_rm", profile={}) == expected
        assert etcd_state.mod_watch("key", sfun="wait_rm_key", profile={}) == expected

        # Test watch with bad sfun
        kwargs = {"sfun": "bad_sfun"}
        expected = {
            "name": "key",
            "changes": {},
            "comment": (
                "etcd.{0[sfun]} does not work with the watch requisite, "
                "please use etcd.wait_set or etcd.wait_rm".format(kwargs)
            ),
            "result": False,
        }
        assert etcd_state.mod_watch("key", **kwargs) == expected
        assert etcd_state.mod_watch("key", **kwargs) == expected



"""
Return data to an etcd server or cluster

:depends: - python-etcd or etcd3-py

In order to return to an etcd server, a profile should be created in the master
configuration file:

.. code-block:: yaml

    my_etcd_config:
      etcd.host: 127.0.0.1
      etcd.port: 2379

It is technically possible to configure etcd without using a profile, but this
is not considered to be a best practice, especially when multiple etcd servers
or clusters are available.

.. code-block:: yaml

    etcd.host: 127.0.0.1
    etcd.port: 2379

In order to choose whether to use etcd API v2 or v3, you can put the following
configuration option in the same place as your etcd configuration.  This option
defaults to true, meaning you will use v2 unless you specify otherwise.

.. code-block:: yaml

    etcd.require_v2: True

When using API v3, there are some specific options available to be configured
within your etcd profile.  They are defaulted to the following...

.. code-block:: yaml

    etcd.encode_keys: False
    etcd.encode_values: True
    etcd.raw_keys: False
    etcd.raw_values: False
    etcd.unicode_errors: "surrogateescape"

``etcd.encode_keys`` indicates whether you want to pre-encode keys using msgpack before
adding them to etcd.

.. note::

    If you set ``etcd.encode_keys`` to ``True``, all recursive functionality will no longer work.
    This includes ``tree`` and ``ls`` and all other methods if you set ``recurse``/``recursive`` to ``True``.
    This is due to the fact that when encoding with msgpack, keys like ``/salt`` and ``/salt/stack`` will have
    differing byte prefixes, and etcd v3 searches recursively using prefixes.

``etcd.encode_values`` indicates whether you want to pre-encode values using msgpack before
adding them to etcd.  This defaults to ``True`` to avoid data loss on non-string values wherever possible.

``etcd.raw_keys`` determines whether you want the raw key or a string returned.

``etcd.raw_values`` determines whether you want the raw value or a string returned.

``etcd.unicode_errors`` determines what you policy to follow when there are encoding/decoding errors.

Additionally, two more options must be specified in the top-level configuration
in order to use the etcd returner:

.. code-block:: yaml

    etcd.returner: my_etcd_config
    etcd.returner_root: /salt/return

The ``etcd.returner`` option specifies which configuration profile to use. The
``etcd.returner_root`` option specifies the path inside etcd to use as the root
of the returner system.

Once the etcd options are configured, the returner may be used:

CLI Example:

    salt '*' test.ping --return etcd

A username and password can be set:

.. code-block:: yaml

    etcd.username: larry  # Optional; requires etcd.password to be set
    etcd.password: 123pass  # Optional; requires etcd.username to be set

You can also set a TTL (time to live) value for the returner:

.. code-block:: yaml

    etcd.ttl: 5

Authentication with username and password, and ttl, currently requires the
``master`` branch of ``python-etcd``.

You may also specify different roles for read and write operations. First,
create the profiles as specified above. Then add:

.. code-block:: yaml

    etcd.returner_read_profile: my_etcd_read
    etcd.returner_write_profile: my_etcd_write
"""

import logging

import salt.utils.jid
import salt.utils.json

try:
    import salt.utils.etcd_util

    if salt.utils.etcd_util.HAS_ETCD_V2 or salt.utils.etcd_util.HAS_ETCD_V3:
        HAS_LIBS = True
    else:
        HAS_LIBS = False
except ImportError:
    HAS_LIBS = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "etcd"


def __virtual__():
    """
    Only return if python-etcd is installed
    """
    if HAS_LIBS:
        return __virtualname__

    return False, "Could not import etcd returner; python-etcd is not installed."


def _get_conn(opts, profile=None):
    """
    Establish a connection to etcd
    """
    if profile is None:
        profile = opts.get("etcd.returner")
    path = opts.get("etcd.returner_root", "/salt/return")
    return salt.utils.etcd_util.get_conn(opts, profile), path


def returner(ret):
    """
    Return data to an etcd server or cluster
    """
    write_profile = __opts__.get("etcd.returner_write_profile")
    if write_profile:
        ttl = __opts__.get(write_profile, {}).get("etcd.ttl")
    else:
        ttl = __opts__.get("etcd.ttl")

    client, path = _get_conn(__opts__, write_profile)
    # Make a note of this minion for the external job cache
    client.set(
        "/".join((path, "minions", ret["id"])),
        ret["jid"],
        ttl=ttl,
    )

    for field in ret:
        # Not using os.path.join because we're not dealing with file paths
        dest = "/".join((path, "jobs", ret["jid"], ret["id"], field))
        client.set(dest, salt.utils.json.dumps(ret[field]), ttl=ttl)


def save_load(jid, load, minions=None):
    """
    Save the load to the specified jid
    """
    log.debug("sdstack_etcd returner <save_load> called jid: %s", jid)
    write_profile = __opts__.get("etcd.returner_write_profile")
    client, path = _get_conn(__opts__, write_profile)
    if write_profile:
        ttl = __opts__.get(write_profile, {}).get("etcd.ttl")
    else:
        ttl = __opts__.get("etcd.ttl")
    client.set(
        "/".join((path, "jobs", jid, ".load.p")),
        salt.utils.json.dumps(load),
        ttl=ttl,
    )


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    """
    Included for API consistency
    """


def clean_old_jobs():
    """
    Included for API consistency
    """


def get_load(jid):
    """
    Return the load data that marks a specified jid
    """
    log.debug("sdstack_etcd returner <get_load> called jid: %s", jid)
    read_profile = __opts__.get("etcd.returner_read_profile")
    client, path = _get_conn(__opts__, read_profile)
    return salt.utils.json.loads(client.get("/".join((path, "jobs", jid, ".load.p"))))


def get_jid(jid):
    """
    Return the information returned when the specified job id was executed
    """
    log.debug("sdstack_etcd returner <get_jid> called jid: %s", jid)
    ret = {}
    client, path = _get_conn(__opts__)
    items = client.get("/".join((path, "jobs", jid)), recurse=True)
    for id, value in items.items():
        if str(id).endswith(".load.p"):
            continue
        id = id.split("/")[-1]
        ret[id] = {"return": salt.utils.json.loads(value["return"])}
    return ret


def get_fun(fun):
    """
    Return a dict of the last function called for all minions
    """
    log.debug("sdstack_etcd returner <get_fun> called fun: %s", fun)
    ret = {}
    client, path = _get_conn(__opts__)
    items = client.get("/".join((path, "minions")), recurse=True)
    for id, jid in items.items():
        id = str(id).rsplit("/", maxsplit=1)[-1]
        efun = salt.utils.json.loads(
            client.get("/".join((path, "jobs", str(jid), id, "fun")))
        )
        if efun == fun:
            ret[id] = str(efun)
    return ret


def get_jids():
    """
    Return a list of all job ids
    """
    log.debug("sdstack_etcd returner <get_jids> called")
    ret = []
    client, path = _get_conn(__opts__)
    items = client.get("/".join((path, "jobs")), recurse=True)
    for key, value in items.items():
        if isinstance(value, dict):  # dict means directory
            jid = str(key).rsplit("/", maxsplit=1)[-1]
            ret.append(jid)
    return ret


def get_minions():
    """
    Return a list of minions
    """
    log.debug("sdstack_etcd returner <get_minions> called")
    ret = []
    client, path = _get_conn(__opts__)
    items = client.get("/".join((path, "minions")), recurse=True)
    for id, _ in items.items():
        id = str(id).rsplit("/", maxsplit=1)[-1]
        ret.append(id)
    return ret


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



"""
Trigger an event in IFTTT
=========================

This state is useful for trigging events in IFTTT.

.. versionadded:: 2015.8.0

.. code-block:: yaml

    ifttt-event:
      ifttt.trigger_event:
        - event: TestEvent
        - value1: 'This state was executed successfully.'
        - value2: 'Another value we can send.'
        - value3: 'A third value we can send.'

The api key can be specified in the master or minion configuration like below:
.. code-block:: yaml

    ifttt:
      secret_key: bzMRb-KKIAaNOwKEEw792J7Eb-B3z7muhdhYblJn4V6

"""


def __virtual__():
    """
    Only load if the ifttt module is available in __salt__
    """
    if "ifttt.trigger_event" in __salt__:
        return "ifttt"
    return (False, "ifttt module could not be loaded")


def trigger_event(name, event, value1=None, value2=None, value3=None):
    """
    Trigger an event in IFTTT

    .. code-block:: yaml

        ifttt-event:
          ifttt.trigger_event:
            - event: TestEvent
            - value1: 'A value that we want to send.'
            - value2: 'A second value that we want to send.'
            - value3: 'A third value that we want to send.'

    The following parameters are required:

    name
        The unique name for this event.

    event
        The name of the event to trigger in IFTTT.

    The following parameters are optional:

    value1
        One of the values that we can send to IFTT.

    value2
        One of the values that we can send to IFTT.

    value3
        One of the values that we can send to IFTT.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if __opts__["test"]:
        ret["comment"] = "The following trigger would be sent to IFTTT: {}".format(
            event
        )
        ret["result"] = None
        return ret

    ret["result"] = __salt__["ifttt.trigger_event"](
        event=event, value1=value1, value2=value2, value3=value3
    )

    if ret and ret["result"]:
        ret["result"] = True
        ret["comment"] = f"Triggered Event: {name}"
    else:
        ret["comment"] = f"Failed to trigger event: {name}"

    return ret



import os
import pathlib
import re
import shutil
import sys
from contextlib import contextmanager

import pytest

import salt.utils.platform
from salt.exceptions import CommandNotFoundError
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.helpers import VirtualEnv, patched_environ

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.requires_network,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
]


@pytest.fixture
def venv(tmp_path):
    with VirtualEnv(venv_dir=tmp_path / "the-venv") as venv:
        yield venv


@pytest.fixture
def pip(modules):
    with patched_environ(
        PIP_SOURCE_DIR="",
        PIP_BUILD_DIR="",
        __cleanup__=[k for k in os.environ if k.startswith("PIP_")],
    ):
        yield modules.pip


def _pip_successful_install(
    target,
    expect=(
        "irc3-plugins-test",
        "pep8",
    ),
):
    """
    Isolate regex for extracting `successful install` message from pip
    """

    expect = set(expect)
    expect_str = "|".join(expect)

    success = re.search(
        r"^.*Successfully installed\s([^\n]+)(?:Clean.*)?", target, re.M | re.S
    )

    success_for = (
        re.findall(rf"({expect_str})(?:-(?:[\d\.-]))?", success.groups()[0])
        if success
        else []
    )

    return expect.issubset(set(success_for))


@pytest.mark.parametrize(
    "pip_version",
    (
        pytest.param(
            "pip==9.0.3",
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 10),
                reason="'pip==9.0.3' is not available on Py >= 3.10",
            ),
        ),
        "pip<20.0",
        "pip<21.0",
        "pip>=21.0",
    ),
)
def test_list_available_packages(pip, pip_version, tmp_path):
    with VirtualEnv(venv_dir=tmp_path, pip_requirement=pip_version) as virtualenv:
        virtualenv.install("-U", pip_version)
        package_name = "pep8"
        available_versions = pip.list_all_versions(
            package_name, bin_env=str(virtualenv.venv_bin_dir)
        )
        assert available_versions


@pytest.mark.parametrize(
    "pip_version",
    (
        "pip==9.0.3",
        "pip<20.0",
        "pip<21.0",
        "pip>=21.0",
    ),
)
def test_list_available_packages_with_index_url(pip, pip_version, tmp_path):
    if sys.version_info < (3, 6) and pip_version == "pip>=21.0":
        pytest.skip(f"{pip_version} is not available on Py3.5")
    if sys.version_info >= (3, 10) and pip_version == "pip==9.0.3":
        pytest.skip(f"{pip_version} is not available on Py3.10")
    with VirtualEnv(venv_dir=tmp_path, pip_requirement=pip_version) as virtualenv:
        virtualenv.install("-U", pip_version)
        package_name = "pep8"
        available_versions = pip.list_all_versions(
            package_name,
            bin_env=str(virtualenv.venv_bin_dir),
            index_url="https://pypi.python.org/simple",
        )
        assert available_versions


def test_issue_2087_missing_pip(venv, pip, modules):
    # Let's remove the pip binary
    pip_bin = venv.venv_bin_dir / "pip"
    site_dir = pathlib.Path(
        modules.virtualenv.get_distribution_path(str(venv.venv_dir), "pip")
    )
    if salt.utils.platform.is_windows():
        pip_bin = venv.venv_dir / "Scripts" / "pip.exe"
        site_dir = venv.venv_dir / "lib" / "site-packages"
    if not pip_bin.is_file():
        pytest.skip("Failed to find the pip binary to the test virtualenv")
    pip_bin.unlink()

    # Also remove the pip dir from site-packages
    # This is needed now that we're using python -m pip instead of the
    # pip binary directly. python -m pip will still work even if the
    # pip binary is missing
    shutil.rmtree(site_dir / "pip")

    with pytest.raises(CommandNotFoundError) as exc:
        pip.freeze(bin_env=str(venv.venv_dir))

    assert str(exc.value) == "Could not find a `pip` binary"

    with pytest.raises(CommandNotFoundError) as exc:
        pip.list(bin_env=str(venv.venv_dir))

    assert str(exc.value) == "Could not find a `pip` binary"


def test_requirements_as_list_of_chains__cwd_set__absolute_file_path(
    venv, pip, tmp_path
):
    # Create a requirements file that depends on another one.
    req1_filename = tmp_path / "requirements1.txt"
    req1_filename.write_text("-r requirements1b.txt\n", encoding="utf-8")
    req1b_filename = tmp_path / "requirements1b.txt"
    req1b_filename.write_text("irc3-plugins-test\n", encoding="utf-8")
    req2_filename = tmp_path / "requirements2.txt"
    req2_filename.write_text("-r requirements2b.txt\n", encoding="utf-8")
    req2b_filename = tmp_path / "requirements2b.txt"
    req2b_filename.write_text("pep8\n", encoding="utf-8")

    ret = pip.install(
        requirements=[str(req1_filename), str(req2_filename)],
        bin_env=venv.venv_dir,
        cwd=tmp_path,
    )
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert _pip_successful_install(ret["stdout"])


def test_requirements_as_list_of_chains__cwd_not_set__absolute_file_path(
    venv, pip, tmp_path
):
    # Create a requirements file that depends on another one.
    req1_filename = tmp_path / "requirements1.txt"
    req1_filename.write_text("-r requirements1b.txt\n", encoding="utf-8")
    req1b_filename = tmp_path / "requirements1b.txt"
    req1b_filename.write_text("irc3-plugins-test\n", encoding="utf-8")
    req2_filename = tmp_path / "requirements2.txt"
    req2_filename.write_text("-r requirements2b.txt\n", encoding="utf-8")
    req2b_filename = tmp_path / "requirements2b.txt"
    req2b_filename.write_text("pep8\n", encoding="utf-8")

    ret = pip.install(
        requirements=[str(req1_filename), str(req2_filename)],
        bin_env=venv.venv_dir,
    )
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert _pip_successful_install(ret["stdout"])


def test_requirements_as_list__absolute_file_path(venv, pip, tmp_path):
    # Create a requirements file that depends on another one.
    req1_filename = tmp_path / "requirements1.txt"
    req1_filename.write_text("irc3-plugins-test\n", encoding="utf-8")
    req2_filename = tmp_path / "requirements2.txt"
    req2_filename.write_text("pep8\n", encoding="utf-8")

    ret = pip.install(
        requirements=[str(req1_filename), str(req2_filename)],
        bin_env=venv.venv_dir,
    )
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert _pip_successful_install(ret["stdout"])


def test_requirements_as_list__non_absolute_file_path(venv, pip, tmp_path):
    # Create a requirements file that depends on another one.
    req1_filename = tmp_path / "requirements1.txt"
    req1_filename.write_text("irc3-plugins-test\n", encoding="utf-8")
    req2_filename = tmp_path / "requirements2.txt"
    req2_filename.write_text("pep8\n", encoding="utf-8")

    ret = pip.install(
        requirements=[str(req1_filename.name), str(req2_filename.name)],
        bin_env=venv.venv_dir,
        cwd=tmp_path,
    )
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert _pip_successful_install(ret["stdout"])


def test_chained_requirements__absolute_file_path(venv, pip, tmp_path):
    # Create a requirements file that depends on another one.
    req1_filename = tmp_path / "requirements1.txt"
    req1_filename.write_text("-r requirements2.txt\n", encoding="utf-8")
    req2_filename = tmp_path / "requirements2.txt"
    req2_filename.write_text("pep8\n", encoding="utf-8")

    ret = pip.install(
        requirements=str(req1_filename),
        bin_env=venv.venv_dir,
    )
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert "installed pep8" in ret["stdout"]


def test_chained_requirements__non_absolute_file_path(venv, pip, tmp_path):
    # Create a requirements file that depends on another one.
    req1_filename = tmp_path / "requirements1.txt"
    req1_filename.write_text("-r requirements2.txt\n", encoding="utf-8")
    req2_filename = tmp_path / "requirements2.txt"
    req2_filename.write_text("pep8\n", encoding="utf-8")

    ret = pip.install(
        requirements=str(req1_filename.name), bin_env=venv.venv_dir, cwd=tmp_path
    )
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert "installed pep8" in ret["stdout"]


@pytest.fixture
def installed_requirement(venv, pip):

    @contextmanager
    def _install(version=None):
        requirement = requirement_name = "pep8"
        if version is not None:
            requirement += f"=={version}"
        ret = pip.install(pkgs=requirement, bin_env=venv.venv_dir)
        assert ret
        assert isinstance(ret, dict)
        assert ret["retcode"] == 0
        assert "installed pep8" in ret["stdout"]
        print(444, pip.freeze(bin_env=venv.venv_dir))
        yield requirement_name

    return _install


def test_pip_uninstall(venv, pip, installed_requirement):
    with installed_requirement() as requirement:
        ret = pip.uninstall(pkgs=requirement, bin_env=venv.venv_dir)
        assert f"uninstalled {requirement}" in ret["stdout"]


def test_pip_install_upgrade(venv, pip, installed_requirement):
    with installed_requirement(version="1.3.4") as requirement:
        ret = pip.install(
            pkgs=requirement,
            bin_env=venv.venv_dir,
            upgrade=True,
        )
        assert f"installed {requirement}" in ret["stdout"]


def test_pip_install_multiple_editables(venv, pip):
    editables = [
        "git+https://github.com/saltstack/istr.git@v1.0.1#egg=iStr",
        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
    ]
    ret = pip.install(editable=editables, bin_env=venv.venv_dir)
    assert _pip_successful_install(ret["stdout"], ("iStr", "SaltTesting"))


def test_pip_install_multiple_editables_and_pkgs(venv, pip):
    editables = [
        "git+https://github.com/saltstack/istr.git@v1.0.1#egg=iStr",
        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
    ]
    ret = pip.install(pkgs="pep8", editable=editables, bin_env=venv.venv_dir)
    assert _pip_successful_install(ret["stdout"], ("iStr", "SaltTesting", "pep8"))


@pytest.mark.parametrize("touch", [True, False])
def test_pip_non_existent_log_file(venv, pip, tmp_path, touch):
    log_file = tmp_path / "tmp-pip-install.log"
    if touch:
        log_file.touch()
    ret = pip.install(pkgs="pep8", log=str(log_file), bin_env=venv.venv_dir)
    assert _pip_successful_install(ret["stdout"], ("pep8",))
    assert log_file.exists()
    assert "pep8" in log_file.read_text()


@pytest.mark.skipif(
    shutil.which("/bin/pip3") is None, reason="Could not find /bin/pip3"
)
@pytest.mark.destructive_test
@pytest.mark.skip_on_windows(reason="test specific for linux usage of /bin/python")
@pytest.mark.skip_initial_gh_actions_failure(
    reason="This was skipped on older golden images and is failing on newer."
)
def test_system_pip3(pip):
    pkg = "lazyimport"
    pkgver = f"{pkg}==0.0.1"
    ret = pip.install(pkgs=pkgver, bin_env="/bin/pip3")
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert f"installed {pkg}" in ret["stdout"]

    ret = pip.freeze(bin_env="/bin/pip3")
    assert pkgver in ret

    ret = pip.uninstall(pkgs=pkg, bin_env="/bin/pip3")
    assert ret
    assert isinstance(ret, dict)
    assert ret["retcode"] == 0
    assert f"uninstalled {pkg}" in ret["stdout"]

    ret = pip.freeze(bin_env="/bin/pip3")
    assert pkg not in ret



"""
Salt interface to LDAP commands

:depends:   - ldap Python module
:configuration: In order to connect to LDAP, certain configuration is required
    in the minion config on the LDAP server. The minimum configuration items
    that must be set are:

    .. code-block:: yaml

        ldap.basedn: dc=acme,dc=com (example values, adjust to suit)

    If your LDAP server requires authentication then you must also set:

    .. code-block:: yaml

        ldap.anonymous: False
        ldap.binddn: admin
        ldap.bindpw: password

    In addition, the following optional values may be set:

    .. code-block:: yaml

        ldap.server: localhost (default=localhost, see warning below)
        ldap.port: 389 (default=389, standard port)
        ldap.tls: False (default=False, no TLS)
        ldap.no_verify: False (default=False, verify TLS)
        ldap.anonymous: True (default=True, bind anonymous)
        ldap.scope: 2 (default=2, ldap.SCOPE_SUBTREE)
        ldap.attrs: [saltAttr] (default=None, return all attributes)

.. warning::

    At the moment this module only recommends connection to LDAP services
    listening on ``localhost``. This is deliberate to avoid the potentially
    dangerous situation of multiple minions sending identical update commands
    to the same LDAP server. It's easy enough to override this behavior, but
    badness may ensue - you have been warned.
"""

import logging
import time

import salt.utils.data
from salt.exceptions import CommandExecutionError

try:
    import ldap
    import ldap.modlist  # pylint: disable=no-name-in-module

    HAS_LDAP = True
except ImportError:
    HAS_LDAP = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "ldap"


def __virtual__():
    """
    Only load this module if the ldap config is set
    """
    # These config items must be set in the minion config
    if HAS_LDAP:
        return __virtualname__
    return (
        False,
        "The ldapmod execution module cannot be loaded: ldap config not present.",
    )


def _config(name, key=None, **kwargs):
    """
    Return a value for 'name' from command line args then config file options.
    Specify 'key' if the config file option is not the same as 'name'.
    """
    if key is None:
        key = name
    if name in kwargs:
        value = kwargs[name]
    else:
        value = __salt__["config.option"](f"ldap.{key}")
    return salt.utils.data.decode(value, to_str=True)


def _connect(**kwargs):
    """
    Instantiate LDAP Connection class and return an LDAP connection object
    """
    connargs = {}
    for name in [
        "uri",
        "server",
        "port",
        "tls",
        "no_verify",
        "binddn",
        "bindpw",
        "anonymous",
    ]:
        connargs[name] = _config(name, **kwargs)

    return _LDAPConnection(**connargs).ldap


def search(
    filter,  # pylint: disable=C0103
    dn=None,  # pylint: disable=C0103
    scope=None,
    attrs=None,
    **kwargs,
):
    """
    Run an arbitrary LDAP query and return the results.

    CLI Example:

    .. code-block:: bash

        salt 'ldaphost' ldap.search "filter=cn=myhost"

    Return data:

    .. code-block:: python

        {'myhost': {'count': 1,
                    'results': [['cn=myhost,ou=hosts,o=acme,c=gb',
                                 {'saltKeyValue': ['ntpserver=ntp.acme.local',
                                                   'foo=myfoo'],
                                  'saltState': ['foo', 'bar']}]],
                    'time': {'human': '1.2ms', 'raw': '0.00123'}}}

    Search and connection options can be overridden by specifying the relevant
    option as key=value pairs, for example:

    .. code-block:: bash

        salt 'ldaphost' ldap.search filter=cn=myhost dn=ou=hosts,o=acme,c=gb
        scope=1 attrs='' server='localhost' port='7393' tls=True bindpw='ssh'
    """
    if not dn:
        dn = _config("dn", "basedn")  # pylint: disable=C0103
    if not scope:
        scope = _config("scope")
    if attrs == "":  # Allow command line 'return all' attr override
        attrs = None
    elif attrs is None:
        attrs = _config("attrs")
    _ldap = _connect(**kwargs)
    start = time.time()
    log.debug(
        "Running LDAP search with filter:%s, dn:%s, scope:%s, attrs:%s",
        filter,
        dn,
        scope,
        attrs,
    )
    results = _ldap.search_s(dn, int(scope), filter, attrs)
    elapsed = time.time() - start
    if elapsed < 0.200:
        elapsed_h = str(round(elapsed * 1000, 1)) + "ms"
    else:
        elapsed_h = str(round(elapsed, 2)) + "s"

    ret = {
        "results": results,
        "count": len(results),
        "time": {"human": elapsed_h, "raw": str(round(elapsed, 5))},
    }
    return ret


class _LDAPConnection:
    """
    Setup an LDAP connection.
    """

    def __init__(self, uri, server, port, tls, no_verify, binddn, bindpw, anonymous):
        """
        Bind to an LDAP directory using passed credentials.
        """
        self.uri = uri
        self.server = server
        self.port = port
        self.tls = tls
        self.binddn = binddn
        self.bindpw = bindpw

        if self.uri == "":
            self.uri = f"ldap://{self.server}:{self.port}"

        try:
            if no_verify:
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

            self.ldap = ldap.initialize(f"{self.uri}")
            self.ldap.protocol_version = 3  # ldap.VERSION3
            self.ldap.set_option(ldap.OPT_REFERRALS, 0)  # Needed for AD

            if self.tls:
                self.ldap.start_tls_s()

            if not anonymous:
                self.ldap.simple_bind_s(self.binddn, self.bindpw)
        except Exception as ldap_error:  # pylint: disable=broad-except
            raise CommandExecutionError(
                "Failed to bind to LDAP server {} as {}: {}".format(
                    self.uri, self.binddn, ldap_error
                )
            )



"""
tests.unit.context_test
~~~~~~~~~~~~~~~~~~~~~~~
"""

import threading
import time

import pytest

import salt.ext.tornado.gen
import salt.ext.tornado.stack_context
import salt.utils.json
from salt.ext.tornado.testing import AsyncTestCase, gen_test
from salt.utils.context import ContextDict, NamespacedDictWrapper
from tests.support.unit import TestCase


class ContextDictTests(AsyncTestCase):
    # how many threads/coroutines to run at a time
    num_concurrent_tasks = 5

    def setUp(self):
        super().setUp()
        self.cd = ContextDict()
        # set a global value
        self.cd["foo"] = "global"

    @pytest.mark.slow_test
    def test_threads(self):
        """Verify that ContextDict overrides properly within threads"""
        rets = []

        def tgt(x, s):
            inner_ret = []
            over = self.cd.clone()

            inner_ret.append(self.cd.get("foo"))
            with over:
                inner_ret.append(over.get("foo"))
                over["foo"] = x
                inner_ret.append(over.get("foo"))
                time.sleep(s)
                inner_ret.append(over.get("foo"))
                rets.append(inner_ret)

        threads = []
        for x in range(0, self.num_concurrent_tasks):
            s = self.num_concurrent_tasks - x
            t = threading.Thread(target=tgt, args=(x, s))
            t.start()
            threads.append(t)

        for t in threads:
            t.join()

        for r in rets:
            self.assertEqual(r[0], r[1])
            self.assertEqual(r[2], r[3])

    @gen_test
    @pytest.mark.slow_test
    def test_coroutines(self):
        """Verify that ContextDict overrides properly within coroutines"""

        @salt.ext.tornado.gen.coroutine
        def secondary_coroutine(over):
            raise salt.ext.tornado.gen.Return(over.get("foo"))

        @salt.ext.tornado.gen.coroutine
        def tgt(x, s, over):
            inner_ret = []
            # first grab the global
            inner_ret.append(self.cd.get("foo"))
            # grab the child's global (should match)
            inner_ret.append(over.get("foo"))
            # override the global
            over["foo"] = x
            inner_ret.append(over.get("foo"))
            # sleep for some time to let other coroutines do this section of code
            yield salt.ext.tornado.gen.sleep(s)
            # get the value of the global again.
            inner_ret.append(over.get("foo"))
            # Call another coroutine to verify that we keep our context
            r = yield secondary_coroutine(over)
            inner_ret.append(r)
            raise salt.ext.tornado.gen.Return(inner_ret)

        futures = []

        for x in range(0, self.num_concurrent_tasks):
            s = self.num_concurrent_tasks - x
            over = self.cd.clone()

            # pylint: disable=cell-var-from-loop
            f = salt.ext.tornado.stack_context.run_with_stack_context(
                salt.ext.tornado.stack_context.StackContext(lambda: over),
                lambda: tgt(x, s / 5.0, over),
            )
            # pylint: enable=cell-var-from-loop
            futures.append(f)

        wait_iterator = salt.ext.tornado.gen.WaitIterator(*futures)
        while not wait_iterator.done():
            r = yield wait_iterator.next()
            self.assertEqual(r[0], r[1])  # verify that the global value remails
            self.assertEqual(r[2], r[3])  # verify that the override sticks locally
            self.assertEqual(
                r[3], r[4]
            )  # verify that the override sticks across coroutines

    def test_basic(self):
        """Test that the contextDict is a dict"""
        # ensure we get the global value
        self.assertEqual(
            dict(self.cd),
            {"foo": "global"},
        )

    def test_override(self):
        over = self.cd.clone()
        over["bar"] = "global"
        self.assertEqual(
            dict(over),
            {"foo": "global", "bar": "global"},
        )
        self.assertEqual(
            dict(self.cd),
            {"foo": "global"},
        )
        with over:
            self.assertEqual(
                dict(over),
                {"foo": "global", "bar": "global"},
            )
            self.assertEqual(
                dict(self.cd),
                {"foo": "global", "bar": "global"},
            )
            over["bar"] = "baz"
            self.assertEqual(
                dict(over),
                {"foo": "global", "bar": "baz"},
            )
            self.assertEqual(
                dict(self.cd),
                {"foo": "global", "bar": "baz"},
            )
        self.assertEqual(
            dict(over),
            {"foo": "global", "bar": "baz"},
        )
        self.assertEqual(
            dict(self.cd),
            {"foo": "global"},
        )

    def test_multiple_contexts(self):
        cds = []
        for x in range(0, 10):
            cds.append(self.cd.clone(bar=x))
        for x, cd in enumerate(cds):
            self.assertNotIn("bar", self.cd)
            with cd:
                self.assertEqual(
                    dict(self.cd),
                    {"bar": x, "foo": "global"},
                )
        self.assertNotIn("bar", self.cd)


class NamespacedDictWrapperTests(TestCase):
    PREFIX = "prefix"

    def setUp(self):
        self._dict = {}

    def test_single_key(self):
        self._dict["prefix"] = {"foo": "bar"}
        w = NamespacedDictWrapper(self._dict, "prefix")
        self.assertEqual(w["foo"], "bar")

    def test_multiple_key(self):
        self._dict["prefix"] = {"foo": {"bar": "baz"}}
        w = NamespacedDictWrapper(self._dict, ("prefix", "foo"))
        self.assertEqual(w["bar"], "baz")

    def test_json_dumps_single_key(self):
        self._dict["prefix"] = {"foo": {"bar": "baz"}}
        w = NamespacedDictWrapper(self._dict, "prefix")
        self.assertEqual(salt.utils.json.dumps(w), '{"foo": {"bar": "baz"}}')

    def test_json_dumps_multiple_key(self):
        self._dict["prefix"] = {"foo": {"bar": "baz"}}
        w = NamespacedDictWrapper(self._dict, ("prefix", "foo"))
        self.assertEqual(salt.utils.json.dumps(w), '{"bar": "baz"}')



"""
Test the win_runas util
"""

from random import randint

import pytest

import salt.modules.win_useradd as win_useradd
import salt.utils.win_runas as win_runas

try:
    import salt.platform.win

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.skipif(HAS_WIN32 is False, reason="Win32 Libraries not available"),
]


@pytest.fixture
def user():
    with pytest.helpers.create_account() as account:
        yield account


@pytest.fixture
def int_user():
    with pytest.helpers.create_account() as account:
        int_name = randint(10000, 99999)
        win_useradd.rename(account.username, int_name)
        account.username = int_name
        yield account


@pytest.mark.parametrize(
    "cmd, expected",
    [
        ("hostname && whoami", "username"),
        ("hostname && echo foo", "foo"),
        ("hostname && python --version", "Python"),
    ],
)
def test_compound_runas(user, cmd, expected):
    if expected == "username":
        expected = user.username
    result = win_runas.runas(
        cmd=salt.platform.win.prepend_cmd("cmd", cmd),
        username=user.username,
        password=user.password,
    )
    assert expected in result["stdout"]


@pytest.mark.parametrize(
    "cmd, expected",
    [
        ("hostname && whoami", "username"),
        ("hostname && echo foo", "foo"),
        ("hostname && python --version", "Python"),
    ],
)
def test_compound_runas_unpriv(user, cmd, expected):
    if expected == "username":
        expected = user.username
    result = win_runas.runas_unpriv(
        cmd=salt.platform.win.prepend_cmd("cmd", cmd),
        username=user.username,
        password=user.password,
    )
    assert expected in result["stdout"]


def test_runas_str_user(user):
    result = win_runas.runas(
        cmd="whoami", username=user.username, password=user.password
    )
    assert user.username in result["stdout"]


def test_runas_int_user(int_user):
    result = win_runas.runas(
        cmd="whoami", username=int(int_user.username), password=int_user.password
    )
    assert str(int_user.username) in result["stdout"]


def test_runas_unpriv_str_user(user):
    result = win_runas.runas_unpriv(
        cmd="whoami", username=user.username, password=user.password
    )
    assert user.username in result["stdout"]


def test_runas_unpriv_int_user(int_user):
    result = win_runas.runas_unpriv(
        cmd="whoami", username=int(int_user.username), password=int_user.password
    )
    assert str(int_user.username) in result["stdout"]



# pylint: skip-file
from __future__ import absolute_import, division, print_function
from wsgiref.validate import validator

from salt.ext.tornado.escape import json_decode
from salt.ext.tornado.test.httpserver_test import TypeCheckHandler
from salt.ext.tornado.testing import AsyncHTTPTestCase
from salt.ext.tornado.web import RequestHandler, Application
from salt.ext.tornado.wsgi import WSGIApplication, WSGIContainer, WSGIAdapter

from salt.ext.tornado.test import httpserver_test
from salt.ext.tornado.test import web_test


class WSGIContainerTest(AsyncHTTPTestCase):
    def wsgi_app(self, environ, start_response):
        status = "200 OK"
        response_headers = [("Content-Type", "text/plain")]
        start_response(status, response_headers)
        return [b"Hello world!"]

    def get_app(self):
        return WSGIContainer(validator(self.wsgi_app))

    def test_simple(self):
        response = self.fetch("/")
        self.assertEqual(response.body, b"Hello world!")


class WSGIApplicationTest(AsyncHTTPTestCase):
    def get_app(self):
        class HelloHandler(RequestHandler):
            def get(self):
                self.write("Hello world!")

        class PathQuotingHandler(RequestHandler):
            def get(self, path):
                self.write(path)

        # It would be better to run the wsgiref server implementation in
        # another thread instead of using our own WSGIContainer, but this
        # fits better in our async testing framework and the wsgiref
        # validator should keep us honest
        return WSGIContainer(validator(WSGIApplication([
            ("/", HelloHandler),
            ("/path/(.*)", PathQuotingHandler),
            ("/typecheck", TypeCheckHandler),
        ])))

    def test_simple(self):
        response = self.fetch("/")
        self.assertEqual(response.body, b"Hello world!")

    def test_path_quoting(self):
        response = self.fetch("/path/foo%20bar%C3%A9")
        self.assertEqual(response.body, u"foo bar\u00e9".encode("utf-8"))

    def test_types(self):
        headers = {"Cookie": "foo=bar"}
        response = self.fetch("/typecheck?foo=bar", headers=headers)
        data = json_decode(response.body)
        self.assertEqual(data, {})

        response = self.fetch("/typecheck", method="POST", body="foo=bar", headers=headers)
        data = json_decode(response.body)
        self.assertEqual(data, {})


# This is kind of hacky, but run some of the HTTPServer and web tests
# through WSGIContainer and WSGIApplication to make sure everything
# survives repeated disassembly and reassembly.
class WSGIConnectionTest(httpserver_test.HTTPConnectionTest):
    def get_app(self):
        return WSGIContainer(validator(WSGIApplication(self.get_handlers())))


def wrap_web_tests_application():
    result = {}
    for cls in web_test.wsgi_safe_tests:
        class WSGIApplicationWrappedTest(cls):  # type: ignore
            def get_app(self):
                self.app = WSGIApplication(self.get_handlers(),
                                           **self.get_app_kwargs())
                return WSGIContainer(validator(self.app))
        result["WSGIApplication_" + cls.__name__] = WSGIApplicationWrappedTest
    return result


globals().update(wrap_web_tests_application())


def wrap_web_tests_adapter():
    result = {}
    for cls in web_test.wsgi_safe_tests:
        class WSGIAdapterWrappedTest(cls):  # type: ignore
            def get_app(self):
                self.app = Application(self.get_handlers(),
                                       **self.get_app_kwargs())
                return WSGIContainer(validator(WSGIAdapter(self.app)))
        result["WSGIAdapter_" + cls.__name__] = WSGIAdapterWrappedTest
    return result


globals().update(wrap_web_tests_adapter())



"""
runit service module
(http://smarden.org/runit)

This module is compatible with the :mod:`service <salt.states.service>` states,
so it can be used to maintain services using the ``provider`` argument:

.. code-block:: yaml

    myservice:
      service:
        - running
        - provider: runit

Provides virtual `service` module on systems using runit as init.


Service management rules (`sv` command):

    service $n is ENABLED   if file SERVICE_DIR/$n/run exists
    service $n is AVAILABLE if ENABLED or if file AVAIL_SVR_DIR/$n/run exists
    service $n is DISABLED  if AVAILABLE but not ENABLED

    SERVICE_DIR/$n is normally a symlink to a AVAIL_SVR_DIR/$n folder


Service auto-start/stop mechanism:

    `sv` (auto)starts/stops service as soon as SERVICE_DIR/<service> is
    created/deleted, both on service creation or a boot time.

    autostart feature is disabled if file SERVICE_DIR/<n>/down exists. This
    does not affect the current's service status (if already running) nor
    manual service management.


Service's alias:

    Service `sva` is an alias of service `svc` when `AVAIL_SVR_DIR/sva` symlinks
    to folder `AVAIL_SVR_DIR/svc`. `svc` can't be enabled if it is already
    enabled through an alias already enabled, since `sv` files are stored in
    folder `SERVICE_DIR/svc/`.

    XBPS package management uses a service's alias to provides service
    alternative(s), such as chrony and openntpd both aliased to ntpd.
"""

import glob
import logging
import os
import time

import salt.utils.files
import salt.utils.path
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


# Function alias to not shadow built-ins.
__func_alias__ = {"reload_": "reload"}

# which dir sv works with
VALID_SERVICE_DIRS = [
    "/service",
    "/var/service",
    "/etc/service",
]
SERVICE_DIR = None
for service_dir in VALID_SERVICE_DIRS:
    if os.path.exists(service_dir):
        SERVICE_DIR = service_dir
        break

# available service directory(ies)
AVAIL_SVR_DIRS = []

# Define the module's virtual name
__virtualname__ = "runit"
__virtual_aliases__ = ("runit",)


def __virtual__():
    """
    Virtual service only on systems using runit as init process (PID 1).
    Otherwise, use this module with the provider mechanism.
    """
    if __grains__.get("init") == "runit":
        if __grains__["os"] == "Void":
            add_svc_avail_path("/etc/sv")
        global __virtualname__
        __virtualname__ = "service"
        return __virtualname__
    if salt.utils.path.which("sv"):
        return __virtualname__
    return (False, "Runit not available.  Please install sv")


def _service_path(name):
    """
    Return SERVICE_DIR+name if possible

    name
        the service's name to work on
    """
    if not SERVICE_DIR:
        raise CommandExecutionError("Could not find service directory.")
    return os.path.join(SERVICE_DIR, name)


# -- states.service  compatible args
def start(name):
    """
    Start service

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.start <service name>
    """
    cmd = f"sv start {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


# -- states.service compatible args
def stop(name):
    """
    Stop service

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.stop <service name>
    """
    cmd = f"sv stop {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


# -- states.service compatible
def reload_(name):
    """
    Reload service

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.reload <service name>
    """
    cmd = f"sv reload {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


# -- states.service compatible
def restart(name):
    """
    Restart service

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.restart <service name>
    """
    cmd = f"sv restart {_service_path(name)}"
    return not __salt__["cmd.retcode"](cmd)


# -- states.service compatible
def full_restart(name):
    """
    Calls runit.restart()

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.full_restart <service name>
    """
    restart(name)


# -- states.service compatible
def status(name, sig=None):
    """
    Return ``True`` if service is running

    name
        the service's name

    sig
        signature to identify with ps

    CLI Example:

    .. code-block:: bash

        salt '*' runit.status <service name>
    """
    if sig:
        # usual way to do by others (debian_service, netbsdservice).
        # XXX probably does not work here (check 'runsv sshd' instead of 'sshd' ?)
        return bool(__salt__["status.pid"](sig))

    svc_path = _service_path(name)
    if not os.path.exists(svc_path):
        # service does not exist
        return False

    # sv return code is not relevant to get a service status.
    # Check its output instead.
    cmd = f"sv status {svc_path}"
    try:
        out = __salt__["cmd.run_stdout"](cmd)
        return out.startswith("run: ")
    except Exception:  # pylint: disable=broad-except
        # sv (as a command) returned an error
        return False


def _is_svc(svc_path):
    """
    Return ``True`` if directory <svc_path> is really a service:
    file <svc_path>/run exists and is executable

    svc_path
        the (absolute) directory to check for compatibility
    """
    run_file = os.path.join(svc_path, "run")
    if (
        os.path.exists(svc_path)
        and os.path.exists(run_file)
        and os.access(run_file, os.X_OK)
    ):
        return True
    return False


def status_autostart(name):
    """
    Return ``True`` if service <name> is autostarted by sv
    (file $service_folder/down does not exist)
    NB: return ``False`` if the service is not enabled.

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.status_autostart <service name>
    """
    return not os.path.exists(os.path.join(_service_path(name), "down"))


def get_svc_broken_path(name="*"):
    """
    Return list of broken path(s) in SERVICE_DIR that match ``name``

    A path is broken if it is a broken symlink or can not be a runit service

    name
        a glob for service name. default is '*'

    CLI Example:

    .. code-block:: bash

        salt '*' runit.get_svc_broken_path <service name>
    """
    if not SERVICE_DIR:
        raise CommandExecutionError("Could not find service directory.")

    ret = set()

    for el in glob.glob(os.path.join(SERVICE_DIR, name)):
        if not _is_svc(el):
            ret.add(el)
    return sorted(ret)


def get_svc_avail_path():
    """
    Return list of paths that may contain available services
    """
    return AVAIL_SVR_DIRS


def add_svc_avail_path(path):
    """
    Add a path that may contain available services.
    Return ``True`` if added (or already present), ``False`` on error.

    path
        directory to add to AVAIL_SVR_DIRS
    """
    if os.path.exists(path):
        if path not in AVAIL_SVR_DIRS:
            AVAIL_SVR_DIRS.append(path)
        return True
    return False


def _get_svc_path(name="*", status=None):
    """
    Return a list of paths to services with ``name`` that have the specified ``status``

    name
        a glob for service name. default is '*'

    status
        None       : all services (no filter, default choice)
        'DISABLED' : available service(s) that is not enabled
        'ENABLED'  : enabled service (whether started on boot or not)
    """

    # This is the core routine to work with services, called by many
    # other functions of this module.
    #
    # The name of a service is the "apparent" folder's name that contains its
    # "run" script. If its "folder" is a symlink, the service is an "alias" of
    # the targeted service.

    if not SERVICE_DIR:
        raise CommandExecutionError("Could not find service directory.")

    # path list of enabled services as /AVAIL_SVR_DIRS/$service,
    # taking care of any service aliases (do not use os.path.realpath()).
    ena = set()
    for el in glob.glob(os.path.join(SERVICE_DIR, name)):
        if _is_svc(el):
            if os.path.islink(el):
                ena.add(os.readlink(el))
            else:
                ena.add(el)
            log.trace("found enabled service path: %s", el)

    if status == "ENABLED":
        return sorted(ena)

    # path list of available services as /AVAIL_SVR_DIRS/$service
    ava = set()
    for d in AVAIL_SVR_DIRS:
        for el in glob.glob(os.path.join(d, name)):
            if _is_svc(el):
                ava.add(el)
                log.trace("found available service path: %s", el)

    if status == "DISABLED":
        # service available but not enabled
        ret = ava.difference(ena)
    else:
        # default: return available services
        ret = ava.union(ena)

    return sorted(ret)


def _get_svc_list(name="*", status=None):
    """
    Return list of services that have the specified service ``status``

    name
        a glob for service name. default is '*'

    status
        None       : all services (no filter, default choice)
        'DISABLED' : available service that is not enabled
        'ENABLED'  : enabled service (whether started on boot or not)
    """
    return sorted(os.path.basename(el) for el in _get_svc_path(name, status))


def get_svc_alias():
    """
    Returns the list of service's name that are aliased and their alias path(s)
    """

    ret = {}
    for d in AVAIL_SVR_DIRS:
        for el in glob.glob(os.path.join(d, "*")):
            if not os.path.islink(el):
                continue
            psvc = os.readlink(el)
            if not os.path.isabs(psvc):
                psvc = os.path.join(d, psvc)
            nsvc = os.path.basename(psvc)
            if nsvc not in ret:
                ret[nsvc] = []
            ret[nsvc].append(el)
    return ret


def available(name):
    """
    Returns ``True`` if the specified service is available, otherwise returns
    ``False``.

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.available <service name>
    """
    return name in _get_svc_list(name)


def missing(name):
    """
    The inverse of runit.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' runit.missing <service name>
    """
    return name not in _get_svc_list(name)


def get_all():
    """
    Return a list of all available services

    CLI Example:

    .. code-block:: bash

        salt '*' runit.get_all
    """
    return _get_svc_list()


def get_enabled():
    """
    Return a list of all enabled services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    return _get_svc_list(status="ENABLED")


def get_disabled():
    """
    Return a list of all disabled services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
    """
    return _get_svc_list(status="DISABLED")


def enabled(name):
    """
    Return ``True`` if the named service is enabled, ``False`` otherwise

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
    """
    # exhaustive check instead of (only) os.path.exists(_service_path(name))
    return name in _get_svc_list(name, "ENABLED")


def disabled(name):
    """
    Return ``True`` if the named service is disabled, ``False``  otherwise

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    # return True for a non-existent service
    return name not in _get_svc_list(name, "ENABLED")


def show(name):
    """
    Show properties of one or more units/jobs or the manager

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' service.show <service name>
    """
    ret = {}
    ret["enabled"] = False
    ret["disabled"] = True
    ret["running"] = False
    ret["service_path"] = None
    ret["autostart"] = False
    ret["command_path"] = None

    ret["available"] = available(name)
    if not ret["available"]:
        return ret

    ret["enabled"] = enabled(name)
    ret["disabled"] = not ret["enabled"]
    ret["running"] = status(name)
    ret["autostart"] = status_autostart(name)
    ret["service_path"] = _get_svc_path(name)[0]
    if ret["service_path"]:
        ret["command_path"] = os.path.join(ret["service_path"], "run")

    # XXX provide info about alias ?

    return ret


def enable(name, start=False, **kwargs):
    """
    Start service ``name`` at boot.
    Returns ``True`` if operation is successful

    name
        the service's name

    start : False
        If ``True``, start the service once enabled.

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <name> [start=True]
    """

    # non-existent service
    if not available(name):
        return False

    # if service is aliased, refuse to enable it
    alias = get_svc_alias()
    if name in alias:
        log.error("This service is aliased, enable its alias instead")
        return False

    # down_file: file that disables sv autostart
    svc_realpath = _get_svc_path(name)[0]
    down_file = os.path.join(svc_realpath, "down")

    # if service already enabled, remove down_file to
    # let service starts on boot (as requested)
    if enabled(name):
        if os.path.exists(down_file):
            try:
                os.unlink(down_file)
            except OSError:
                log.error("Unable to remove file %s", down_file)
                return False
        return True

    # let's enable the service

    if not start:
        # create a temp 'down' file BEFORE enabling service.
        # will prevent sv from starting this service automatically.
        log.trace("need a temporary file %s", down_file)
        if not os.path.exists(down_file):
            try:
                # pylint: disable=resource-leakage
                salt.utils.files.fopen(down_file, "w").close()
                # pylint: enable=resource-leakage
            except OSError:
                log.error("Unable to create file %s", down_file)
                return False

    # enable the service
    try:
        os.symlink(svc_realpath, _service_path(name))

    except OSError:
        # (attempt to) remove temp down_file anyway
        log.error("Unable to create symlink %s", down_file)
        if not start:
            os.unlink(down_file)
        return False

    # ensure sv is aware of this new service before continuing.
    # if not, down_file might be removed too quickly,
    # before 'sv' have time to take care about it.
    # Documentation indicates that a change is handled within 5 seconds.
    cmd = f"sv status {_service_path(name)}"
    retcode_sv = 1
    count_sv = 0
    while retcode_sv != 0 and count_sv < 10:
        time.sleep(0.5)
        count_sv += 1
        call = __salt__["cmd.run_all"](cmd)
        retcode_sv = call["retcode"]

    # remove the temp down_file in any case.
    if (not start) and os.path.exists(down_file):
        try:
            os.unlink(down_file)
        except OSError:
            log.error("Unable to remove temp file %s", down_file)
            retcode_sv = 1

    # if an error happened, revert our changes
    if retcode_sv != 0:
        os.unlink(os.path.join([_service_path(name), name]))
        return False
    return True


def disable(name, stop=False, **kwargs):
    """
    Don't start service ``name`` at boot
    Returns ``True`` if operation is successful

    name
        the service's name

    stop
        if True, also stops the service

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <name> [stop=True]
    """

    # non-existent as registrered service
    if not enabled(name):
        return False

    # down_file: file that prevent sv autostart
    svc_realpath = _get_svc_path(name)[0]
    down_file = os.path.join(svc_realpath, "down")

    if stop:
        stop(name)

    if not os.path.exists(down_file):
        try:
            # pylint: disable=resource-leakage
            salt.utils.files.fopen(down_file, "w").close()
            # pylint: enable=resource-leakage
        except OSError:
            log.error("Unable to create file %s", down_file)
            return False

    return True


def remove(name):
    """
    Remove the service <name> from system.
    Returns ``True`` if operation is successful.
    The service will be also stopped.

    name
        the service's name

    CLI Example:

    .. code-block:: bash

        salt '*' service.remove <name>
    """

    if not enabled(name):
        return False

    svc_path = _service_path(name)
    if not os.path.islink(svc_path):
        log.error("%s is not a symlink: not removed", svc_path)
        return False

    if not stop(name):
        log.error("Failed to stop service %s", name)
        return False
    try:
        os.remove(svc_path)
    except OSError:
        log.error("Unable to remove symlink %s", svc_path)
        return False
    return True



import logging
import random
import string

import pytest

import salt.config
import salt.loader
import salt.modules.boto_cloudwatch_event as boto_cloudwatch_event
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

# pylint: disable=import-error,no-name-in-module,unused-import
try:
    import boto
    import boto3
    from botocore import __version__ as found_botocore_version
    from botocore.exceptions import ClientError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
]

# pylint: enable=import-error,no-name-in-module,unused-import
log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    else:
        return True


if _has_required_boto():
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    not_found_error = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
                "Message": "Test-defined error",
            }
        },
        "msg",
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    rule_name = "test_thing_type"
    rule_desc = "test_thing_type_desc"
    rule_sched = "rate(20 min)"
    rule_arn = "arn:::::rule/arn"
    rule_ret = dict(
        Arn=rule_arn,
        Description=rule_desc,
        EventPattern=None,
        Name=rule_name,
        RoleArn=None,
        ScheduleExpression=rule_sched,
        State="ENABLED",
    )
    create_rule_ret = dict(
        Name=rule_name,
    )
    target_ret = dict(
        Id="target1",
    )


class BotoCloudWatchEventTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto3", "args", "systemd", "path", "platform"], context={}
        )
        return {boto_cloudwatch_event: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto_cloudwatch_event.__init__(self.opts)
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        session_instance.client.return_value = self.conn


class BotoCloudWatchEventTestCaseMixin:
    pass


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
class BotoCloudWatchEventTestCase(
    BotoCloudWatchEventTestCaseBase, BotoCloudWatchEventTestCaseMixin
):
    """
    TestCase for salt.modules.boto_cloudwatch_event module
    """

    def test_that_when_checking_if_a_rule_exists_and_a_rule_exists_the_rule_exists_method_returns_true(
        self,
    ):
        """
        Tests checking event existence when the event already exists
        """
        self.conn.list_rules.return_value = {"Rules": [rule_ret]}
        result = boto_cloudwatch_event.exists(Name=rule_name, **conn_parameters)

        self.assertTrue(result["exists"])

    def test_that_when_checking_if_a_rule_exists_and_a_rule_does_not_exist_the_exists_method_returns_false(
        self,
    ):
        """
        Tests checking rule existence when the rule does not exist
        """
        self.conn.list_rules.return_value = {"Rules": []}
        result = boto_cloudwatch_event.exists(Name=rule_name, **conn_parameters)

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_a_rule_exists_and_boto3_returns_an_error_the_rule_exists_method_returns_error(
        self,
    ):
        """
        Tests checking rule existence when boto returns an error
        """
        self.conn.list_rules.side_effect = ClientError(error_content, "list_rules")
        result = boto_cloudwatch_event.exists(Name=rule_name, **conn_parameters)

        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("list_rules")
        )

    def test_that_when_describing_rule_and_rule_exists_the_describe_rule_method_returns_rule(
        self,
    ):
        """
        Tests describe rule for an existing rule
        """
        self.conn.describe_rule.return_value = rule_ret
        result = boto_cloudwatch_event.describe(Name=rule_name, **conn_parameters)

        self.assertEqual(result.get("rule"), rule_ret)

    def test_that_when_describing_rule_and_rule_does_not_exists_the_describe_method_returns_none(
        self,
    ):
        """
        Tests describe rule for an non existent rule
        """
        self.conn.describe_rule.side_effect = not_found_error
        result = boto_cloudwatch_event.describe(Name=rule_name, **conn_parameters)

        self.assertNotEqual(result.get("error"), None)

    def test_that_when_describing_rule_and_boto3_returns_error_the_describe_method_returns_error(
        self,
    ):
        self.conn.describe_rule.side_effect = ClientError(
            error_content, "describe_rule"
        )
        result = boto_cloudwatch_event.describe(Name=rule_name, **conn_parameters)

        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("describe_rule"),
        )

    def test_that_when_creating_a_rule_succeeds_the_create_rule_method_returns_true(
        self,
    ):
        """
        tests True when rule created
        """
        self.conn.put_rule.return_value = create_rule_ret
        result = boto_cloudwatch_event.create_or_update(
            Name=rule_name,
            Description=rule_desc,
            ScheduleExpression=rule_sched,
            **conn_parameters
        )
        self.assertTrue(result["created"])

    def test_that_when_creating_a_rule_fails_the_create_method_returns_error(self):
        """
        tests False when rule not created
        """
        self.conn.put_rule.side_effect = ClientError(error_content, "put_rule")
        result = boto_cloudwatch_event.create_or_update(
            Name=rule_name,
            Description=rule_desc,
            ScheduleExpression=rule_sched,
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("put_rule")
        )

    def test_that_when_deleting_a_rule_succeeds_the_delete_method_returns_true(self):
        """
        tests True when delete rule succeeds
        """
        self.conn.delete_rule.return_value = {}
        result = boto_cloudwatch_event.delete(Name=rule_name, **conn_parameters)

        self.assertTrue(result.get("deleted"))
        self.assertEqual(result.get("error"), None)

    def test_that_when_deleting_a_rule_fails_the_delete_method_returns_error(self):
        """
        tests False when delete rule fails
        """
        self.conn.delete_rule.side_effect = ClientError(error_content, "delete_rule")
        result = boto_cloudwatch_event.delete(Name=rule_name, **conn_parameters)
        self.assertFalse(result.get("deleted"))
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("delete_rule")
        )

    def test_that_when_listing_targets_and_rule_exists_the_list_targets_method_returns_targets(
        self,
    ):
        """
        Tests listing targets for an existing rule
        """
        self.conn.list_targets_by_rule.return_value = {"Targets": [target_ret]}
        result = boto_cloudwatch_event.list_targets(Rule=rule_name, **conn_parameters)

        self.assertEqual(result.get("targets"), [target_ret])

    def test_that_when_listing_targets_and_rule_does_not_exist_the_list_targets_method_returns_error(
        self,
    ):
        """
        Tests list targets for an non existent rule
        """
        self.conn.list_targets_by_rule.side_effect = not_found_error
        result = boto_cloudwatch_event.list_targets(Rule=rule_name, **conn_parameters)

        self.assertNotEqual(result.get("error"), None)

    def test_that_when_putting_targets_succeeds_the_put_target_method_returns_no_failures(
        self,
    ):
        """
        tests None when targets added
        """
        self.conn.put_targets.return_value = {"FailedEntryCount": 0}
        result = boto_cloudwatch_event.put_targets(
            Rule=rule_name, Targets=[], **conn_parameters
        )
        self.assertIsNone(result["failures"])

    def test_that_when_putting_targets_fails_the_put_targets_method_returns_error(self):
        """
        tests False when thing type not created
        """
        self.conn.put_targets.side_effect = ClientError(error_content, "put_targets")
        result = boto_cloudwatch_event.put_targets(
            Rule=rule_name, Targets=[], **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("put_targets")
        )

    def test_that_when_removing_targets_succeeds_the_remove_targets_method_returns_true(
        self,
    ):
        """
        tests True when remove targets succeeds
        """
        self.conn.remove_targets.return_value = {"FailedEntryCount": 0}
        result = boto_cloudwatch_event.remove_targets(
            Rule=rule_name, Ids=[], **conn_parameters
        )

        self.assertIsNone(result["failures"])
        self.assertEqual(result.get("error"), None)

    def test_that_when_removing_targets_fails_the_remove_targets_method_returns_error(
        self,
    ):
        """
        tests False when remove targets fails
        """
        self.conn.remove_targets.side_effect = ClientError(
            error_content, "remove_targets"
        )
        result = boto_cloudwatch_event.remove_targets(
            Rule=rule_name, Ids=[], **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("remove_targets"),
        )



import pytest

from salt.modules import boto_ssm
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():

    return {
        boto_ssm: {
            "__utils__": {
                "boto3.assign_funcs": MagicMock(),
            },
        }
    }


def test___virtual_has_boto_reqs_true():
    with patch("salt.utils.versions.check_boto_reqs", return_value=True):
        result = boto_ssm.__virtual__()
    assert result is True


def test___virtual_has_boto_reqs_false():
    with patch("salt.utils.versions.check_boto_reqs", return_value=False):
        result = boto_ssm.__virtual__()
    assert result is False



import pytest

import salt.minion


def test_minion_send_req_async(minion_opts):
    """
    Ensure Minion._send_req_sync raises a TimeoutError when no reply is
    received.
    """

    minion = salt.minion.Minion(minion_opts)
    with pytest.raises(TimeoutError):
        minion._send_req_sync({"load": "meh"}, 10)



import logging
import textwrap

import pytest

import salt.modules.cmdmod as cmdmod
import salt.modules.file as filemod
from tests.support.mock import MagicMock, Mock, patch

log = logging.getLogger(__name__)

pytestmark = pytest.mark.skip_on_windows(
    reason="Chattr shouldn't be available on Windows"
)


@pytest.fixture
def _common_patches():
    with patch("salt.utils.platform.is_aix", Mock(return_value=False)), patch(
        "os.path.exists", Mock(return_value=True)
    ), patch("salt.utils.path.which", Mock(return_value="some/tune2fs")):
        yield


@pytest.fixture
def configure_loader_modules(_common_patches):
    return {
        filemod: {
            "__salt__": {"cmd.run": cmdmod.run},
            "__opts__": {"test": False},
        },
    }


def test_chattr_version_returns_None_if_no_tune2fs_exists():
    with patch("salt.utils.path.which", Mock(return_value="")):
        actual = filemod._chattr_version()
        assert actual is None


def test_on_aix_chattr_version_should_be_None_even_if_tune2fs_exists():
    patch_which = patch(
        "salt.utils.path.which",
        Mock(return_value="fnord"),
    )
    patch_aix = patch(
        "salt.utils.platform.is_aix",
        Mock(return_value=True),
    )
    mock_run = MagicMock(return_value="fnord")
    patch_run = patch.dict(filemod.__salt__, {"cmd.run": mock_run})
    with patch_which, patch_aix, patch_run:
        actual = filemod._chattr_version()
        assert actual is None
        mock_run.assert_not_called()


def test_chattr_version_should_return_version_from_tune2fs():
    expected = "1.43.4"
    sample_output = textwrap.dedent(
        """
        tune2fs 1.43.4 (31-Jan-2017)
        Usage: tune2fs [-c max_mounts_count] [-e errors_behavior] [-f] [-g group]
        [-i interval[d|m|w]] [-j] [-J journal_options] [-l]
        [-m reserved_blocks_percent] [-o [^]mount_options[,...]]
        [-p mmp_update_interval] [-r reserved_blocks_count] [-u user]
        [-C mount_count] [-L volume_label] [-M last_mounted_dir]
        [-O [^]feature[,...]] [-Q quota_options]
        [-E extended-option[,...]] [-T last_check_time] [-U UUID]
        [-I new_inode_size] [-z undo_file] device
        """
    )
    patch_which = patch(
        "salt.utils.path.which",
        Mock(return_value="fnord"),
    )
    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": MagicMock(return_value=sample_output)},
    )
    with patch_which, patch_run:
        actual = filemod._chattr_version()
        assert actual == expected


def test_if_tune2fs_has_no_version_version_should_be_None():
    patch_which = patch(
        "salt.utils.path.which",
        Mock(return_value="fnord"),
    )
    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": MagicMock(return_value="fnord")},
    )
    with patch_which, patch_run:
        actual = filemod._chattr_version()
        assert actual is None


def test_chattr_has_extended_attrs_should_return_False_if_chattr_version_is_None():
    patch_chattr = patch(
        "salt.modules.file._chattr_version",
        Mock(return_value=None),
    )
    with patch_chattr:
        actual = filemod._chattr_has_extended_attrs()
        assert not actual, actual


def test_chattr_has_extended_attrs_should_return_False_if_version_is_too_low():
    below_expected = "0.1.1"
    patch_chattr = patch(
        "salt.modules.file._chattr_version",
        Mock(return_value=below_expected),
    )
    with patch_chattr:
        actual = filemod._chattr_has_extended_attrs()
        assert not actual, actual


def test_chattr_has_extended_attrs_should_return_False_if_version_is_equal_threshold():
    threshold = "1.41.12"
    patch_chattr = patch(
        "salt.modules.file._chattr_version",
        Mock(return_value=threshold),
    )
    with patch_chattr:
        actual = filemod._chattr_has_extended_attrs()
        assert not actual, actual


def test_chattr_has_extended_attrs_should_return_True_if_version_is_above_threshold():
    higher_than = "1.41.13"
    patch_chattr = patch(
        "salt.modules.file._chattr_version",
        Mock(return_value=higher_than),
    )
    with patch_chattr:
        actual = filemod._chattr_has_extended_attrs()
        assert actual, actual


def test_check_perms_should_report_no_attr_changes_if_there_are_none():
    filename = "/path/to/fnord"
    attrs = "aAcCdDeijPsStTu"

    higher_than = "1.41.13"
    patch_chattr = patch(
        "salt.modules.file._chattr_version",
        Mock(return_value=higher_than),
    )
    patch_exists = patch(
        "os.path.exists",
        Mock(return_value=True),
    )
    patch_stats = patch(
        "salt.modules.file.stats",
        Mock(return_value={"user": "foo", "group": "bar", "mode": "123"}),
    )
    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": MagicMock(return_value="--------- " + filename)},
    )
    with patch_chattr, patch_exists, patch_stats, patch_run:
        actual_ret, actual_perms = filemod.check_perms(
            name=filename,
            ret=None,
            user="foo",
            group="bar",
            mode="123",
            attrs=attrs,
            follow_symlinks=False,
        )
        assert actual_ret.get("changes", {}).get("attrs") is None, actual_ret


def test_check_perms_should_report_attrs_new_and_old_if_they_changed():
    filename = "/path/to/fnord"
    attrs = "aAcCdDeijPsStTu"
    existing_attrs = "aeiu"
    expected = {
        "attrs": {"old": existing_attrs, "new": attrs},
    }

    higher_than = "1.41.13"
    patch_chattr = patch(
        "salt.modules.file._chattr_version",
        Mock(return_value=higher_than),
    )
    patch_stats = patch(
        "salt.modules.file.stats",
        Mock(return_value={"user": "foo", "group": "bar", "mode": "123"}),
    )
    patch_cmp = patch(
        "salt.modules.file._cmp_attrs",
        MagicMock(
            side_effect=[
                filemod.AttrChanges(
                    added="aAcCdDeijPsStTu",
                    removed="",
                ),
                filemod.AttrChanges(
                    None,
                    None,
                ),
            ]
        ),
    )
    patch_chattr = patch(
        "salt.modules.file.chattr",
        MagicMock(),
    )

    def fake_cmd(cmd, *args, **kwargs):
        if cmd == ["lsattr", "/path/to/fnord"]:
            return textwrap.dedent(
                """
                {}---- {}
                """.format(
                    existing_attrs, filename
                )
            ).strip()
        else:
            assert False, f"not sure how to handle {cmd}"

    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": MagicMock(side_effect=fake_cmd)},
    )
    patch_ver = patch(
        "salt.modules.file._chattr_has_extended_attrs",
        MagicMock(return_value=True),
    )
    with patch_chattr, patch_stats, patch_cmp, patch_run, patch_ver:
        actual_ret, actual_perms = filemod.check_perms(
            name=filename,
            ret=None,
            user="foo",
            group="bar",
            mode="123",
            attrs=attrs,
            follow_symlinks=False,
        )
        assert actual_ret["changes"] == expected



"""
Module to import docker-compose via saltstack

.. versionadded:: 2016.3.0

:maintainer: Jean Praloran <jeanpralo@gmail.com>
:maturity: new
:depends: docker-compose>=1.5
:platform: all

Introduction
------------
This module allows one to deal with docker-compose file in a directory.

This is  a first version only, the following commands are missing at the moment
but will be built later on if the community is interested in this module:

- run
- logs
- port
- scale

Installation Prerequisites
--------------------------

This execution module requires at least version 1.4.0 of both docker-compose_ and
Docker_. docker-compose can easily be installed using :py:func:`pip.install
<salt.modules.pip.install>`:

.. code-block:: bash

    salt myminion pip.install docker-compose>=1.5.0

.. _docker-compose: https://pypi.python.org/pypi/docker-compose
.. _Docker: https://www.docker.com/


How to use this module?
-----------------------
In order to use the module if you have no docker-compose file on the server you
can issue the command create, it takes two arguments the path where the
docker-compose.yml will be stored and the content of this latter:

.. code-block:: text

    # salt-call -l debug dockercompose.create /tmp/toto '
    database:
    image: mongo:3.0
    command: mongod --smallfiles --quiet --logpath=/dev/null
    '

Then you can execute a list of method defined at the bottom with at least one
argument (the path where the docker-compose.yml will be read) and an optional
python list which corresponds to the services names:

.. code-block:: bash

    # salt-call -l debug dockercompose.up /tmp/toto
    # salt-call -l debug dockercompose.restart /tmp/toto '[database]'
    # salt-call -l debug dockercompose.stop /tmp/toto
    # salt-call -l debug dockercompose.rm /tmp/toto

Docker-compose method supported
-------------------------------
- up
- restart
- stop
- start
- pause
- unpause
- kill
- rm
- ps
- pull
- build

Functions
---------
- docker-compose.yml management
    - :py:func:`dockercompose.create <salt.modules.dockercompose.create>`
    - :py:func:`dockercompose.get <salt.modules.dockercompose.get>`
- Manage containers
    - :py:func:`dockercompose.restart <salt.modules.dockercompose.restart>`
    - :py:func:`dockercompose.stop <salt.modules.dockercompose.stop>`
    - :py:func:`dockercompose.pause <salt.modules.dockercompose.pause>`
    - :py:func:`dockercompose.unpause <salt.modules.dockercompose.unpause>`
    - :py:func:`dockercompose.start <salt.modules.dockercompose.start>`
    - :py:func:`dockercompose.kill <salt.modules.dockercompose.kill>`
    - :py:func:`dockercompose.rm <salt.modules.dockercompose.rm>`
    - :py:func:`dockercompose.up <salt.modules.dockercompose.up>`
- Manage containers image:
    - :py:func:`dockercompose.pull <salt.modules.dockercompose.pull>`
    - :py:func:`dockercompose.build <salt.modules.dockercompose.build>`
- Gather information about containers:
    - :py:func:`dockercompose.ps <salt.modules.dockercompose.ps>`
- Manage service definitions:
    - :py:func:`dockercompose.service_create <salt.modules.dockercompose.ps>`
    - :py:func:`dockercompose.service_upsert <salt.modules.dockercompose.ps>`
    - :py:func:`dockercompose.service_remove <salt.modules.dockercompose.ps>`
    - :py:func:`dockercompose.service_set_tag <salt.modules.dockercompose.ps>`

Detailed Function Documentation
-------------------------------
"""

import inspect
import logging
import os
import re
from operator import attrgetter

import salt.utils.files
import salt.utils.stringutils
from salt.serializers import json
from salt.utils import yaml

try:
    import compose
    from compose.cli.command import get_project
    from compose.service import ConvergenceStrategy

    HAS_DOCKERCOMPOSE = True
except ImportError:
    HAS_DOCKERCOMPOSE = False

try:
    from compose.project import OneOffFilter

    USE_FILTERCLASS = True
except ImportError:
    USE_FILTERCLASS = False

MIN_DOCKERCOMPOSE = (1, 5, 0)
VERSION_RE = r"([\d.]+)"

log = logging.getLogger(__name__)
debug = False

__virtualname__ = "dockercompose"
DEFAULT_DC_FILENAMES = ("docker-compose.yml", "docker-compose.yaml")


def __virtual__():
    if HAS_DOCKERCOMPOSE:
        match = re.match(VERSION_RE, str(compose.__version__))
        if match:
            version = tuple(int(x) for x in match.group(1).split("."))
            if version >= MIN_DOCKERCOMPOSE:
                return __virtualname__
    return (
        False,
        "The dockercompose execution module not loaded: "
        "compose python library not available.",
    )


def __standardize_result(status, message, data=None, debug_msg=None):
    """
    Standardizes all responses

    :param status:
    :param message:
    :param data:
    :param debug_msg:
    :return:
    """
    result = {"status": status, "message": message}

    if data is not None:
        result["return"] = data

    if debug_msg is not None and debug:
        result["debug"] = debug_msg

    return result


def __get_docker_file_path(path):
    """
    Determines the filepath to use

    :param path:
    :return:
    """
    if os.path.isfile(path):
        return path
    for dc_filename in DEFAULT_DC_FILENAMES:
        file_path = os.path.join(path, dc_filename)
        if os.path.isfile(file_path):
            return file_path
    # implicitly return None


def __read_docker_compose_file(file_path):
    """
    Read the compose file if it exists in the directory

    :param file_path:
    :return:
    """
    if not os.path.isfile(file_path):
        return __standardize_result(
            False, f"Path {file_path} is not present", None, None
        )
    try:
        with salt.utils.files.fopen(file_path, "r") as fl:
            file_name = os.path.basename(file_path)
            result = {file_name: ""}
            for line in fl:
                result[file_name] += salt.utils.stringutils.to_unicode(line)
    except OSError:
        return __standardize_result(False, f"Could not read {file_path}", None, None)
    return __standardize_result(True, f"Reading content of {file_path}", result, None)


def __load_docker_compose(path):
    """
    Read the compose file and load its' contents

    :param path:
    :return:
    """
    file_path = __get_docker_file_path(path)
    if file_path is None:
        msg = f"Could not find docker-compose file at {path}"
        return None, __standardize_result(False, msg, None, None)
    if not os.path.isfile(file_path):
        return (
            None,
            __standardize_result(False, f"Path {file_path} is not present", None, None),
        )
    try:
        with salt.utils.files.fopen(file_path, "r") as fl:
            loaded = yaml.load(fl)
    except OSError:
        return (
            None,
            __standardize_result(False, f"Could not read {file_path}", None, None),
        )
    except yaml.YAMLError as yerr:
        msg = f"Could not parse {file_path} {yerr}"
        return None, __standardize_result(False, msg, None, None)
    if not loaded:
        msg = f"Got empty compose file at {file_path}"
        return None, __standardize_result(False, msg, None, None)
    if "services" not in loaded:
        loaded["services"] = {}
    result = {"compose_content": loaded, "file_name": os.path.basename(file_path)}
    return result, None


def __dump_docker_compose(path, content, already_existed):
    """
    Dumps

    :param path:
    :param content: the not-yet dumped content
    :return:
    """
    try:
        dumped = yaml.safe_dump(content, indent=2, default_flow_style=False)
        return __write_docker_compose(path, dumped, already_existed)
    except TypeError as t_err:
        msg = f"Could not dump {content} {t_err}"
        return __standardize_result(False, msg, None, None)


def __write_docker_compose(path, docker_compose, already_existed):
    """
    Write docker-compose to a path
    in order to use it with docker-compose ( config check )

    :param path:

    docker_compose
        contains the docker-compose file

    :return:
    """
    if path.lower().endswith((".yml", ".yaml")):
        file_path = path
        dir_name = os.path.dirname(path)
    else:
        dir_name = path
        file_path = os.path.join(dir_name, DEFAULT_DC_FILENAMES[0])
    if os.path.isdir(dir_name) is False:
        os.mkdir(dir_name)
    try:
        with salt.utils.files.fopen(file_path, "w") as fl:
            fl.write(salt.utils.stringutils.to_str(docker_compose))
    except OSError:
        return __standardize_result(False, f"Could not write {file_path}", None, None)
    project = __load_project_from_file_path(file_path)
    if isinstance(project, dict):
        if not already_existed:
            os.remove(file_path)
        return project
    return file_path


def __load_project(path):
    """
    Load a docker-compose project from path

    :param path:
    :return:
    """
    file_path = __get_docker_file_path(path)
    if file_path is None:
        msg = f"Could not find docker-compose file at {path}"
        return __standardize_result(False, msg, None, None)
    return __load_project_from_file_path(file_path)


def __load_project_from_file_path(file_path):
    """
    Load a docker-compose project from file path

    :param path:
    :return:
    """
    try:
        project = get_project(
            project_dir=os.path.dirname(file_path),
            config_path=[os.path.basename(file_path)],
        )
    except Exception as inst:  # pylint: disable=broad-except
        return __handle_except(inst)
    return project


def __load_compose_definitions(path, definition):
    """
    Will load the compose file located at path
    Then determines the format/contents of the sent definition

    err or results are only set if there were any

    :param path:
    :param definition:
    :return tuple(compose_result, loaded_definition, err):
    """
    compose_result, err = __load_docker_compose(path)
    if err:
        return None, None, err
    if isinstance(definition, dict):
        return compose_result, definition, None
    elif definition.strip().startswith("{"):
        try:
            loaded_definition = json.deserialize(definition)
        except json.DeserializationError as jerr:
            msg = f"Could not parse {definition} {jerr}"
            return None, None, __standardize_result(False, msg, None, None)
    else:
        try:
            loaded_definition = yaml.load(definition)
        except yaml.YAMLError as yerr:
            msg = f"Could not parse {definition} {yerr}"
            return None, None, __standardize_result(False, msg, None, None)
    return compose_result, loaded_definition, None


def __dump_compose_file(path, compose_result, success_msg, already_existed):
    """
    Utility function to dump the compose result to a file.

    :param path:
    :param compose_result:
    :param success_msg: the message to give upon success
    :return:
    """
    ret = __dump_docker_compose(
        path, compose_result["compose_content"], already_existed
    )
    if isinstance(ret, dict):
        return ret
    return __standardize_result(
        True, success_msg, compose_result["compose_content"], None
    )


def __handle_except(inst):
    """
    Handle exception and return a standard result

    :param inst:
    :return:
    """
    return __standardize_result(
        False,
        f"Docker-compose command {inspect.stack()[1][3]} failed",
        f"{inst}",
        None,
    )


def _get_convergence_plans(project, service_names):
    """
    Get action executed for each container

    :param project:
    :param service_names:
    :return:
    """
    ret = {}
    plans = project._get_convergence_plans(
        project.get_services(service_names), ConvergenceStrategy.changed
    )
    for cont in plans:
        (action, container) = plans[cont]
        if action == "create":
            ret[cont] = "Creating container"
        elif action == "recreate":
            ret[cont] = "Re-creating container"
        elif action == "start":
            ret[cont] = "Starting container"
        elif action == "noop":
            ret[cont] = "Container is up to date"
    return ret


def get(path):
    """
    Get the content of the docker-compose file into a directory

    path
        Path where the docker-compose file is stored on the server

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.get /path/where/docker-compose/stored
    """
    file_path = __get_docker_file_path(path)
    if file_path is None:
        return __standardize_result(False, f"Path {path} is not present", None, None)
    salt_result = __read_docker_compose_file(file_path)
    if not salt_result["status"]:
        return salt_result
    project = __load_project(path)
    if isinstance(project, dict):
        salt_result["return"]["valid"] = False
    else:
        salt_result["return"]["valid"] = True
    return salt_result


def create(path, docker_compose):
    """
    Create and validate a docker-compose file into a directory

    path
        Path where the docker-compose file will be stored on the server

    docker_compose
        docker_compose file

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.create /path/where/docker-compose/stored content
    """
    if docker_compose:
        ret = __write_docker_compose(path, docker_compose, already_existed=False)
        if isinstance(ret, dict):
            return ret
    else:
        return __standardize_result(
            False,
            "Creating a docker-compose project failed, you must send a valid"
            " docker-compose file",
            None,
            None,
        )
    return __standardize_result(
        True,
        "Successfully created the docker-compose file",
        {"compose.base_dir": path},
        None,
    )


def pull(path, service_names=None):
    """
    Pull image for containers in the docker-compose file, service_names is a
    python list, if omitted pull all images

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will pull only the image for the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.pull /path/where/docker-compose/stored
        salt myminion dockercompose.pull /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    if isinstance(project, dict):
        return project
    else:
        try:
            project.pull(service_names)
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Pulling containers images via docker-compose succeeded", None, None
    )


def build(path, service_names=None):
    """
    Build image for containers in the docker-compose file, service_names is a
    python list, if omitted build images for all containers. Please note
    that at the moment the module does not allow you to upload your Dockerfile,
    nor any other file you could need with your docker-compose.yml, you will
    have to make sure the files you need are actually in the directory specified
    in the `build` keyword

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will pull only the image for the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.build /path/where/docker-compose/stored
        salt myminion dockercompose.build /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    if isinstance(project, dict):
        return project
    else:
        try:
            project.build(service_names)
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Building containers images via docker-compose succeeded", None, None
    )


def restart(path, service_names=None):
    """
    Restart container(s) in the docker-compose file, service_names is a python
    list, if omitted restart all containers

    path
        Path where the docker-compose file is stored on the server

    service_names
        If specified will restart only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.restart /path/where/docker-compose/stored
        salt myminion dockercompose.restart /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    debug_ret = {}
    result = {}
    if isinstance(project, dict):
        return project
    else:
        try:
            project.restart(service_names)
            if debug:
                for container in project.containers():
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
                        result[container.get("Name")] = "restarted"
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Restarting containers via docker-compose", result, debug_ret
    )


def stop(path, service_names=None):
    """
    Stop running containers in the docker-compose file, service_names is a python
    list, if omitted stop all containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will stop only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.stop /path/where/docker-compose/stored
        salt myminion dockercompose.stop  /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    debug_ret = {}
    result = {}
    if isinstance(project, dict):
        return project
    else:
        try:
            project.stop(service_names)
            if debug:
                for container in project.containers(stopped=True):
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
                        result[container.get("Name")] = "stopped"
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Stopping containers via docker-compose", result, debug_ret
    )


def pause(path, service_names=None):
    """
    Pause running containers in the docker-compose file, service_names is a python
    list, if omitted pause all containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will pause only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.pause /path/where/docker-compose/stored
        salt myminion dockercompose.pause /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    debug_ret = {}
    result = {}
    if isinstance(project, dict):
        return project
    else:
        try:
            project.pause(service_names)
            if debug:
                for container in project.containers():
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
                        result[container.get("Name")] = "paused"
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Pausing containers via docker-compose", result, debug_ret
    )


def unpause(path, service_names=None):
    """
    Un-Pause containers in the docker-compose file, service_names is a python
    list, if omitted unpause all containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will un-pause only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.pause /path/where/docker-compose/stored
        salt myminion dockercompose.pause /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    debug_ret = {}
    result = {}
    if isinstance(project, dict):
        return project
    else:
        try:
            project.unpause(service_names)
            if debug:
                for container in project.containers():
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
                        result[container.get("Name")] = "unpaused"
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Un-Pausing containers via docker-compose", result, debug_ret
    )


def start(path, service_names=None):
    """
    Start containers in the docker-compose file, service_names is a python
    list, if omitted start all containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will start only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.start /path/where/docker-compose/stored
        salt myminion dockercompose.start /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    debug_ret = {}
    result = {}
    if isinstance(project, dict):
        return project
    else:
        try:
            project.start(service_names)
            if debug:
                for container in project.containers():
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
                        result[container.get("Name")] = "started"
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Starting containers via docker-compose", result, debug_ret
    )


def kill(path, service_names=None):
    """
    Kill containers in the docker-compose file, service_names is a python
    list, if omitted kill all containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will kill only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.kill /path/where/docker-compose/stored
        salt myminion dockercompose.kill /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    debug_ret = {}
    result = {}
    if isinstance(project, dict):
        return project
    else:
        try:
            project.kill(service_names)
            if debug:
                for container in project.containers(stopped=True):
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
                        result[container.get("Name")] = "killed"
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Killing containers via docker-compose", result, debug_ret
    )


def rm(path, service_names=None):
    """
    Remove stopped containers in the docker-compose file, service_names is a python
    list, if omitted remove all stopped containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will remove only the specified stopped services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.rm /path/where/docker-compose/stored
        salt myminion dockercompose.rm /path/where/docker-compose/stored '[janus]'
    """

    project = __load_project(path)
    if isinstance(project, dict):
        return project
    else:
        try:
            project.remove_stopped(service_names)
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Removing stopped containers via docker-compose", None, None
    )


def ps(path):
    """
    List all running containers and report some information about them

    path
        Path where the docker-compose file is stored on the server

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.ps /path/where/docker-compose/stored
    """

    project = __load_project(path)
    result = {}
    if isinstance(project, dict):
        return project
    else:
        if USE_FILTERCLASS:
            containers = sorted(
                project.containers(None, stopped=True)
                + project.containers(None, OneOffFilter.only),
                key=attrgetter("name"),
            )
        else:
            containers = sorted(
                project.containers(None, stopped=True)
                + project.containers(None, one_off=True),
                key=attrgetter("name"),
            )
        for container in containers:
            command = container.human_readable_command
            if len(command) > 30:
                command = f"{command[:26]} ..."
            result[container.name] = {
                "id": container.id,
                "name": container.name,
                "command": command,
                "state": container.human_readable_state,
                "ports": container.human_readable_ports,
            }
    return __standardize_result(True, "Listing docker-compose containers", result, None)


def up(path, service_names=None):
    """
    Create and start containers defined in the docker-compose.yml file
    located in path, service_names is a python list, if omitted create and
    start all containers

    path
        Path where the docker-compose file is stored on the server
    service_names
        If specified will create and start only the specified services

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.up /path/where/docker-compose/stored
        salt myminion dockercompose.up /path/where/docker-compose/stored '[janus]'
    """

    debug_ret = {}
    project = __load_project(path)
    if isinstance(project, dict):
        return project
    else:
        try:
            result = _get_convergence_plans(project, service_names)
            ret = project.up(service_names)
            if debug:
                for container in ret:
                    if (
                        service_names is None
                        or container.get("Name")[1:] in service_names
                    ):
                        container.inspect_if_not_inspected()
                        debug_ret[container.get("Name")] = container.inspect()
        except Exception as inst:  # pylint: disable=broad-except
            return __handle_except(inst)
    return __standardize_result(
        True, "Adding containers via docker-compose", result, debug_ret
    )


def service_create(path, service_name, definition):
    """
    Create the definition of a docker-compose service
    This fails when the service already exists
    This does not pull or up the service
    This wil re-write your yaml file. Comments will be lost. Indentation is set to 2 spaces

    path
        Path where the docker-compose file is stored on the server
    service_name
        Name of the service to create
    definition
        Service definition as yaml or json string

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.service_create /path/where/docker-compose/stored service_name definition
    """
    compose_result, loaded_definition, err = __load_compose_definitions(
        path, definition
    )
    if err:
        return err
    services = compose_result["compose_content"]["services"]
    if service_name in services:
        msg = f"Service {service_name} already exists"
        return __standardize_result(False, msg, None, None)
    services[service_name] = loaded_definition
    return __dump_compose_file(
        path,
        compose_result,
        f"Service {service_name} created",
        already_existed=True,
    )


def service_upsert(path, service_name, definition):
    """
    Create or update the definition of a docker-compose service
    This does not pull or up the service
    This wil re-write your yaml file. Comments will be lost. Indentation is set to 2 spaces

    path
        Path where the docker-compose file is stored on the server
    service_name
        Name of the service to create
    definition
        Service definition as yaml or json string

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.service_upsert /path/where/docker-compose/stored service_name definition
    """
    compose_result, loaded_definition, err = __load_compose_definitions(
        path, definition
    )
    if err:
        return err
    services = compose_result["compose_content"]["services"]
    if service_name in services:
        msg = f"Service {service_name} already exists"
        return __standardize_result(False, msg, None, None)
    services[service_name] = loaded_definition
    return __dump_compose_file(
        path,
        compose_result,
        f"Service definition for {service_name} is set",
        already_existed=True,
    )


def service_remove(path, service_name):
    """
    Remove the definition of a docker-compose service
    This does not rm the container
    This wil re-write your yaml file. Comments will be lost. Indentation is set to 2 spaces

    path
        Path where the docker-compose file is stored on the server
    service_name
        Name of the service to remove

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.service_remove /path/where/docker-compose/stored service_name
    """
    compose_result, err = __load_docker_compose(path)
    if err:
        return err
    services = compose_result["compose_content"]["services"]
    if service_name not in services:
        return __standardize_result(
            False, f"Service {service_name} did not exists", None, None
        )
    del services[service_name]
    return __dump_compose_file(
        path,
        compose_result,
        f"Service {service_name} is removed from {path}",
        already_existed=True,
    )


def service_set_tag(path, service_name, tag):
    """
    Change the tag of a docker-compose service
    This does not pull or up the service
    This wil re-write your yaml file. Comments will be lost. Indentation is set to 2 spaces

    path
        Path where the docker-compose file is stored on the server
    service_name
        Name of the service to remove
    tag
        Name of the tag (often used as version) that the service image should have

    CLI Example:

    .. code-block:: bash

        salt myminion dockercompose.service_create /path/where/docker-compose/stored service_name tag
    """
    compose_result, err = __load_docker_compose(path)
    if err:
        return err
    services = compose_result["compose_content"]["services"]
    if service_name not in services:
        return __standardize_result(
            False, f"Service {service_name} did not exists", None, None
        )
    if "image" not in services[service_name]:
        return __standardize_result(
            False,
            f'Service {service_name} did not contain the variable "image"',
            None,
            None,
        )
    image = services[service_name]["image"].split(":")[0]
    services[service_name]["image"] = f"{image}:{tag}"
    return __dump_compose_file(
        path,
        compose_result,
        f'Service {service_name} is set to tag "{tag}"',
        already_existed=True,
    )



"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""

import copy
import os
import pathlib
import shutil
import sys
import textwrap

import pytest

import salt.fileclient
import salt.fileserver.roots as roots
import salt.utils.files
import salt.utils.hashutils
import salt.utils.platform
import salt.utils.stringutils
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture(scope="module")
def unicode_filename():
    return "питон.txt"


@pytest.fixture(scope="module")
def unicode_dirname():
    return "соль"


@pytest.fixture
def testfile(tmp_path):
    fp = tmp_path / "testfile"
    fp.write_text("This is a testfile")
    return fp


@pytest.fixture
def tmp_state_tree(tmp_path, testfile, unicode_filename, unicode_dirname):
    dirname = tmp_path / "roots_tmp_state_tree"
    dirname.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(str(testfile), str(dirname / testfile.name))
    unicode_dir = dirname / unicode_dirname
    unicode_dir.mkdir(parents=True, exist_ok=True)
    (dirname / unicode_filename).write_text("this is a unicode file")
    (unicode_dir / unicode_filename).write_text(
        "this is a unicode file in a unicode env"
    )
    (unicode_dir / "notunicode").write_text(
        "this is NOT a unicode file in a unicode env"
    )

    return dirname


@pytest.fixture
def testfilepath(tmp_state_tree, testfile):
    return tmp_state_tree / testfile.name


@pytest.fixture
def configure_loader_modules(tmp_state_tree, master_opts):
    overrides = {"file_roots": {"base": [str(tmp_state_tree)]}}
    master_opts.update(overrides)
    return {roots: {"__opts__": master_opts}}


def test_file_list(unicode_filename):
    ret = roots.file_list({"saltenv": "base"})
    assert "testfile" in ret
    assert unicode_filename in ret


def test_find_file(tmp_state_tree):
    ret = roots.find_file("testfile")
    assert "testfile" == ret["rel"]

    full_path_to_file = str(tmp_state_tree / "testfile")
    assert full_path_to_file == ret["path"]


def test_serve_file(testfilepath):
    with patch.dict(roots.__opts__, {"file_buffer_size": 262144}):
        load = {
            "saltenv": "base",
            "path": str(testfilepath),
            "loc": 0,
        }
        fnd = {"path": str(testfilepath), "rel": "testfile"}
        ret = roots.serve_file(load, fnd)

        with salt.utils.files.fopen(str(testfilepath), "rb") as fp_:
            data = fp_.read()

        assert ret == {"data": data, "dest": "testfile"}


def test_envs(unicode_dirname):
    opts = {"file_roots": copy.copy(roots.__opts__["file_roots"])}
    opts["file_roots"][unicode_dirname] = opts["file_roots"]["base"]
    with patch.dict(roots.__opts__, opts):
        ret = roots.envs()
    assert "base" in ret
    assert unicode_dirname in ret


def test_file_hash(testfile):
    load = {
        "saltenv": "base",
        "path": str(testfile),
    }
    fnd = {"path": str(testfile), "rel": "testfile"}
    ret = roots.file_hash(load, fnd)

    # Hashes are different in Windows. May be how git translates line
    # endings
    with salt.utils.files.fopen(str(testfile), "rb") as fp_:
        hsum = salt.utils.hashutils.sha256_digest(fp_.read())

    assert ret == {"hsum": hsum, "hash_type": "sha256"}


def test_file_list_emptydirs(tmp_state_tree):
    empty_dir = tmp_state_tree / "empty_dir"
    empty_dir.mkdir(parents=True, exist_ok=True)
    ret = roots.file_list_emptydirs({"saltenv": "base"})
    assert "empty_dir" in ret


def test_file_list_with_slash(unicode_filename):
    opts = {"file_roots": copy.copy(roots.__opts__["file_roots"])}
    opts["file_roots"]["foo/bar"] = opts["file_roots"]["base"]
    load = {
        "saltenv": "foo/bar",
    }
    with patch.dict(roots.__opts__, opts):
        ret = roots.file_list(load)
    assert "testfile" in ret
    assert unicode_filename in ret


def test_dir_list(tmp_state_tree, unicode_dirname):
    empty_dir = tmp_state_tree / "empty_dir"
    empty_dir.mkdir(parents=True, exist_ok=True)
    ret = roots.dir_list({"saltenv": "base"})
    assert "empty_dir" in ret
    assert unicode_dirname in ret


def test_symlink_list(tmp_state_tree):
    source_sym = tmp_state_tree / "source_sym"
    source_sym.write_text("")
    dest_sym = tmp_state_tree / "dest_sym"
    dest_sym.symlink_to(str(source_sym))
    ret = roots.symlink_list({"saltenv": "base"})
    assert ret == {"dest_sym": str(source_sym)}


def test_dynamic_file_roots(tmp_path):
    dyn_root_dir = tmp_path / "dyn_root_dir"
    dyn_root_dir.mkdir(parents=True, exist_ok=True)
    top_sls = dyn_root_dir / "top.sls"
    with salt.utils.files.fopen(str(top_sls), "w") as fp_:
        fp_.write("{{saltenv}}:\n  '*':\n    - dynamo\n")
    dynamo_sls = dyn_root_dir / "dynamo.sls"
    with salt.utils.files.fopen(str(dynamo_sls), "w") as fp_:
        fp_.write("foo:\n  test.nop\n")
    opts = {"file_roots": copy.copy(roots.__opts__["file_roots"])}
    opts["file_roots"]["__env__"] = [str(dyn_root_dir)]
    with patch.dict(roots.__opts__, opts):
        ret1 = roots.find_file("dynamo.sls", "dyn")
        ret2 = roots.file_list({"saltenv": "dyn"})
    assert "dynamo.sls" == ret1["rel"]
    assert "top.sls" in ret2
    assert "dynamo.sls" in ret2


@pytest.mark.skip_on_windows(
    reason="Windows does not support this master function",
)
def test_update_no_change():
    # process all changes that have happen
    # changes will always take place the first time during testing
    ret = roots.update()
    assert ret["changed"] is True

    # check if no changes took place
    ret = roots.update()
    assert ret["changed"] is False
    assert ret["files"]["changed"] == []
    assert ret["files"]["removed"] == []
    assert ret["files"]["added"] == []


def test_update_mtime_map():
    """
    Test that files with colons in the filename are properly handled in the
    mtime_map, and that they are properly identified as having changed.
    """
    mtime_map_path = pathlib.Path(roots.__opts__["cachedir"], "roots", "mtime_map")
    mtime_map_mock = mock_open(
        read_data={
            str(mtime_map_path): textwrap.dedent(
                """\
                /srv/salt/kleine_Datei.txt:1594263154.0469685
                /srv/salt/große:Datei.txt:1594263160.9336357
                """
            ),
        }
    )
    new_mtime_map = {
        "/srv/salt/kleine_Datei.txt": 1594263154.0469685,
        "/srv/salt/große:Datei.txt": 1594263261.0616212,
    }

    with patch(
        "salt.fileserver.reap_fileserver_cache_dir", MagicMock(return_value=True)
    ), patch(
        "salt.fileserver.generate_mtime_map", MagicMock(return_value=new_mtime_map)
    ), patch.dict(
        roots.__opts__, {"fileserver_events": False}
    ), patch(
        "salt.utils.files.fopen", mtime_map_mock
    ):
        ret = roots.update()

    # Confirm the expected return from the function
    assert ret == {
        "changed": True,
        "files": {
            "changed": ["/srv/salt/große:Datei.txt"],
            "removed": [],
            "added": [],
        },
        "backend": "roots",
    }, ret

    # Confirm that the new values were written to the mtime_map. Sort both
    # lists of lines to account for variances in dictionary iteration order
    # between Python releases.
    lines_written = sorted(mtime_map_mock.write_calls())
    expected = sorted(
        salt.utils.stringutils.to_bytes(f"{key}:{val}\n")
        for key, val in new_mtime_map.items()
    )
    assert lines_written == expected, lines_written


def test_update_mtime_map_unicode_error(tmp_path):
    """
    Test that a malformed mtime_map (which causes an UnicodeDecodeError
    exception) is handled properly.
    """
    new_mtime_map = {
        "/srv/salt/große:Datei.txt": 1594263261.0616212,
    }
    tmpdirname = tmp_path / "unicode_error"
    mtime_map_path = tmpdirname / "roots" / "mtime_map"
    mtime_map_path.parent.mkdir(parents=True, exist_ok=True)
    with salt.utils.files.fopen(str(mtime_map_path), "wb") as fp:
        fp.write(b"\x9c")

    with patch(
        "salt.fileserver.reap_fileserver_cache_dir",
        MagicMock(return_value=True),
    ), patch(
        "salt.fileserver.generate_mtime_map",
        MagicMock(return_value=new_mtime_map),
    ), patch.dict(
        roots.__opts__,
        {"fileserver_events": False, "cachedir": str(tmpdirname)},
    ):
        ret = roots.update()

    assert ret == {
        "changed": True,
        "files": {
            "changed": [],
            "removed": [],
            "added": ["/srv/salt/große:Datei.txt"],
        },
        "backend": "roots",
    }


def test_find_file_not_in_root(tmp_state_tree):
    """
    Fileroots should never 'find' a file that is outside of it's root.
    """
    badfile = pathlib.Path(tmp_state_tree).parent / "bar"
    badfile.write_text("Bad file")
    badpath = "../bar"
    ret = roots.find_file(badpath)
    assert ret == {"path": "", "rel": ""}
    badpath = f"{tmp_state_tree / '..' / 'bar'}"
    ret = roots.find_file(badpath)
    assert ret == {"path": "", "rel": ""}


def test_serve_file_not_in_root(tmp_state_tree):
    """
    Fileroots should never 'serve' a file that is outside of it's root.
    """
    badfile = pathlib.Path(tmp_state_tree).parent / "bar"
    badfile.write_text("Bad file")
    badpath = "../bar"
    load = {"path": "salt://|..\\bar", "saltenv": "base", "loc": 0}
    fnd = {
        "path": f"{tmp_state_tree / '..' / 'bar'}",
        "rel": f"{pathlib.Path('..') / 'bar'}",
    }
    ret = roots.serve_file(load, fnd)
    if "win" in sys.platform:
        assert ret == {"data": "", "dest": "..\\bar"}
    else:
        assert ret == {"data": "", "dest": "../bar"}


def test_find_file_symlink_destination_not_in_root(tmp_state_tree):
    dirname = pathlib.Path(tmp_state_tree).parent / "foo"
    dirname.mkdir(parents=True, exist_ok=True)
    testfile = dirname / "testfile"
    testfile.write_text("testfile")
    symlink = tmp_state_tree / "bar"
    symlink.symlink_to(str(dirname))
    ret = roots.find_file("bar/testfile")
    assert ret["path"] == str(symlink / "testfile")
    assert ret["rel"] == f"bar{os.sep}testfile"


def test_serve_file_symlink_destination_not_in_root(tmp_state_tree):
    dirname = pathlib.Path(tmp_state_tree).parent / "foo"
    dirname.mkdir(parents=True, exist_ok=True)
    testfile = dirname / "testfile"
    testfile.write_text("testfile")
    symlink = tmp_state_tree / "bar"
    symlink.symlink_to(str(dirname))
    load = {"path": "bar/testfile", "saltenv": "base", "loc": 0}
    fnd = {"path": str(symlink / "testfile"), "rel": "bar/testfile"}
    ret = roots.serve_file(load, fnd)
    assert ret == {"data": b"testfile", "dest": "bar/testfile"}


def test_relative_file_roots(tmp_state_tree):
    parent = pathlib.Path(tmp_state_tree).parent
    reldir = os.path.basename(tmp_state_tree)
    opts = {"file_roots": copy.copy(roots.__opts__["file_roots"])}
    opts["file_roots"]["base"] = [reldir]
    with patch.dict(roots.__opts__, opts), pytest.helpers.change_cwd(str(parent)):
        ret = roots.find_file("testfile")
        assert "testfile" == ret["rel"]
